
let swap (a: 'a array) (i: int) (j: int) : unit =
  let tmp = a.(i) in
  a.(i) <- a.(j);
  a.(j) <- tmp

(* mélange de Knuth *)
let shuffle (a: 'a array) : unit =
  for i = 1 to Array.length a - 1 do
    swap a i (Random.int (i + 1))
  done

(* choisir `k` éléments dans un tableau `a` *)
let sampling (k: int) (a: 'a array) : 'a array =
  if k < 0 || k > Array.length a then invalid_arg "sampling";
  let r = Array.sub a 0 k in
  for i = k to Array.length a - 1 do
    let j = Random.int (i + 1) in
    if j < k then r.(j) <- a.(i)
  done;
  r

let insertion_sort a =
  for i = 1 to Array.length a - 1 do
    let v = a.(i) and j = ref i in
    while 0 < !j && v < a.(!j - 1) do
      a.(!j) <- a.(!j - 1);
      decr j
    done;
    a.(!j) <- v
  done

(* le plus petit entier naturel qui n'apparaît pas dans `a` *)
let mex a =
  let n = Array.length a in
  let free = Array.make (n+1) true in
  Array.iter (fun v -> if 0 <= v && v < n then free.(v) <- false) a;
  let rec find i = if i = n || free.(i) then i else find (i+1) in
  find 0

let histogram a =
  let h = Hashtbl.create (Array.length a) in
  let incr v =
    Hashtbl.replace h v (1 + try Hashtbl.find h v with Not_found -> 0) in
  Array.iter incr a;
  let l = Hashtbl.fold (fun v n acc -> (v, n) :: acc) h [] in
  List.sort (fun (v1, _) (v2, _) -> compare v1 v2) l

let print elt fmt a =
  let n = Array.length a in
  Format.fprintf fmt "[|@[";
  for i = 0 to n - 1 do
    elt fmt a.(i);
    if i < n - 1 then Format.fprintf fmt ";@ "
  done;
  Format.fprintf fmt "@]|]"

(* `dicho_mat m v` recherche `v` dans la matrice `m`, en supposant
   que `m` est croissante selon les lignes et selon les colonnes *)
let dicho_mat (m: int array array) (v: int) : bool =
  let rec aux i j k l =
    (* cherche dans [i..j[ x [k..l[ *)
    i < j && k < l && (
    let p = i + (j - i) / 2 in
    let q = k + (l - k) / 2 in
    m.(p).(q) = v ||
    if v < m.(p).(q) then aux i p k     l || aux i     j k q
                     else aux i j (q+1) l || aux (p+1) j k l)
  in
  aux 0 (Array.length m) 0 (Array.length m.(0))

let peak_naive (a: int array) : int =
  let n = Array.length a
  and p = ref 0 in
  for i = 1 to n - 1 do
    if a.(i - 1) < a.(i) then p := i
  done;
  !p

let peak (a: int array) : int =
  let rec aux i j = (* on cherche dans a[i..j[ *)
  if i == j - 1 then
    i
  else
    let k = i + (j - i) / 2 in (* i < k < j *)
    if a.(k-1) < a.(k) then aux k j else aux i k
  in
  aux 0 (Array.length a)

(* entrée : 0 <= l <= k < r <= |a|
   sortie : i avec l <= i <= k et

            l       i  k           r
           +-------+-----+--------+
           |  < p  | = p |  >= p  |
           +-------+-----+--------+ *)
let rec split cmp l k r a =
  assert (l <= k && k < r);
  if r = l + 1 then l else (
  let p = a.(l) in
  let lo = ref (l + 1) in
  let hi = ref r in
  (*  l        lo    hi      r
     +-+------+-----+-------+
     |p|  < p | ??? | >= p  |
     +-+------+-----+-------+ *)
  while !lo < !hi do
    if cmp a.(!lo) p >= 0 then (decr hi; swap a !lo !hi)
                          else incr lo
  done;
  let ip = !lo - 1 in
  swap a l ip;
  (*  l      ip lo=hi      r
     +------+--+----------+
     | < p  |p |   >= p   |
     +------+--+----------+ *)
  if k = ip then k else
  if k < ip then split cmp l k ip a else
  let i = split cmp !lo k r a in
  if i = !lo && a.(i) = p then ip else i
  )
