module Stack = Istack (* nos piles *)

(* Algorithme sur les automates *)

(** acceptation de la chaine [s] par l'automate [a] avec backtracking *)
let accept_d ?(debug = false) a s =
  let count = ref 0 in
  let n = String.length s in
  let rec loop i q =
    if debug then incr count;
    if i = n then Auto.is_final a q
    else
      exists (loop i) (Auto.eps_trans a q)
      || exists (loop (i + 1)) (Auto.trans a q s.[i])

  and exists f l =
    match l with [] -> false | [ q ] -> f q | q :: ll -> f q || exists f ll
  in
  let res = loop 0 0 in
  if debug then (
    incr count;
    Format.eprintf " %d & %!" !count);
  res

let accept = accept_d ~debug:false

(** Formulation impérative *)
let accept_imp a s =
  let todo = Stack.create () in
  let res = ref false in
  let n = String.length s in
  Stack.push todo (0, 0);
  while not (Stack.is_empty todo) do
    let q, i = Stack.pop todo in
    if i = n && Auto.is_final a q then res := true
    else (
      List.iter (fun q2 -> Stack.push todo (q2, i)) (Auto.eps_trans a q);
      List.iter (fun q2 -> Stack.push todo (q2, i + 1)) (Auto.trans a q s.[i]))
  done;
  !res

(* List des états non-accessible *)
let array_elements a =
  let t_elem = ref [] and f_elem = ref [] in
  for q = Array.length a - 1 downto 0 do
    if a.(q) then t_elem := q :: !t_elem else f_elem := q :: !f_elem
  done;
  (!t_elem, !f_elem)

let rec union (l1 : Auto.state list) l2 =
  match (l1, l2) with
  | [], _ -> l2
  | _, [] -> l1
  | p1 :: ll1, p2 :: ll2 ->
      if p1 < p2 then p1 :: union ll1 l2
      else if p1 = p2 then p1 :: union ll1 ll2
      else p2 :: union l1 ll2

module By_iter = struct
  let iter_dfs succ post init =
    let rec loop elem =
      List.iter
        (fun elem2 ->
          loop elem2;
          post elem elem2)
        (succ elem)
    in
    loop init

  let forward auto =
    let visited = Array.make (Auto.size auto) false in
    let succ q =
      if visited.(q) then []
      else (
        visited.(q) <- true;
        List.map snd (Auto.all_trans_opt auto q))
    in
    iter_dfs succ (fun _ _ -> ()) 0;
    array_elements visited

  let backward auto =
    let n = Auto.size auto in
    let rev_auto = Array.make n [] in
    let set_pred q (_, p) = rev_auto.(p) <- q :: rev_auto.(p) in
    let roots = ref [] in
    for q = 0 to n - 1 do
      if Auto.is_final auto q then roots := q :: !roots;
      let trs = Auto.all_trans_opt auto q in
      List.iter (set_pred q) trs
    done;
    let visited = Array.make (Auto.size auto) false in
    let succ q =
      if visited.(q) then []
      else (
        visited.(q) <- true;
        rev_auto.(q))
    in
    for i = 0 to n - 1 do
      if Auto.is_final auto i then iter_dfs succ (fun _ _ -> ()) i
    done;
    array_elements visited

  let eclosure auto =
    let visited = Array.make (Auto.size auto) [] in
    let succ q =
      match visited.(q) with
      | [] ->
          visited.(q) <- [ q ];
          Auto.eps_trans auto q
      | _ -> []
    in
    let post q p = visited.(q) <- union visited.(q) visited.(p) in
    iter_dfs succ post 0;
    visited

  let inter auto1 auto2 =
    let trans = Hashtbl.create 16 in
    let new_id =
      let c = ref ~-1 in
       fun () -> incr c; !c
    in
    let add_trans p c q =
      let id, lst =
        try Hashtbl.find trans p with Not_found -> (new_id (), [])
      in
      Hashtbl.replace trans p (id, (c, q) :: lst)
    in
    let succ ((q1, q2) as q) =
      if Hashtbl.mem trans q then []
      else
        let trs1 = Auto.all_trans_opt auto1 q1 in
        let trs2 = Auto.all_trans_opt auto2 q2 in
        List.fold_left
          (fun acc (c1, p1) ->
            List.fold_left
              (fun acc (c2, p2) ->
                if c1 = c2 then (
                  add_trans q c1 (p1, p2);
                  (p1, p2) :: acc)
                else acc)
              acc trs2)
          [] trs1
    in
    Hashtbl.add trans (0, 0) (new_id (), []);
    iter_dfs succ (fun _ _ -> ()) (0, 0);
    let nauto = Auto.create (Hashtbl.length trans) in
    Hashtbl.iter
      (fun (q1, q2) (id, trs) ->
        if Auto.is_final auto1 q1 && Auto.is_final auto2 q2 then
          Auto.set_final nauto id;
        List.iter
          (fun (c, p) ->
            Auto.add_trans_opt nauto id c (fst (Hashtbl.find trans p)))
          trs)
      trans;
    nauto

    let rem_epsilon auto =
      let ecl = eclosure auto in
      let ltrs = ref [] in
      let nauto = Auto.create (Auto.size auto) in
      for q = 0 to Auto.size auto - 1 do
        for ic = 0 to 255 do
          let c = Char.chr ic in
          List.iter (fun p ->
            let trs = Auto.trans auto p c in
            List.iter (fun s -> ltrs := (q,c,s) :: !ltrs) trs
            ) ecl.(q)
        done;
        if Auto.is_final auto q then Auto.set_final nauto q;
      done;
      List.iter (fun (q,c,p) -> Auto.add_trans nauto q c p) !ltrs;
      nauto
      
end

let eclosure a =
  let visited = Array.make (Auto.size a) [] in
  let rec visit q =
    match visited.(q) with
    | [] ->
        visited.(q) <- [ q ];
        let etrans = Auto.eps_trans a q in
        List.iter
          (fun p ->
            visit p;
            visited.(q) <- union visited.(q) visited.(p))
          etrans
    | _ -> ()
  in
  for q = 0 to Auto.size a - 1 do
    visit q
  done;
  visited

let exists_i f a = Math.exists 0 (Array.length a) (fun i -> f i a.(i))

let nfa_accept a s =
  assert (not (Auto.has_epsilon a));
  let n = String.length s in
  let na = Auto.size a in
  let rec loop qset i other_set =
    if i = n then exists_i (fun q b -> b && Auto.is_final a q) qset
    else
      let found = ref false in
      Array.fill other_set 0 na false;
      for q = 0 to na - 1 do
        if qset.(q) then
          List.iter
            (fun p ->
              found := true;
              other_set.(p) <- true)
            (Auto.trans a q s.[i])
      done;
      !found && loop other_set (i + 1) qset
  in
  let qset = Array.make na false in
  let other_set = Array.make na false in
  qset.(0) <- true;
  loop qset 0 other_set

let enfa_accept a s =
  let n = String.length s in
  let eclosure = eclosure a in
  let na = Auto.size a in
  let mkset () = Array.make na false in
  let rec loop qset i other_set =
    if i = n then exists_i (fun q b -> b && Auto.is_final a q) qset
    else (
      for q = 0 to na - 1 do
        other_set.(q) <- false;
        let b = qset.(q) in
        if b then
          List.iter
            (fun p -> List.iter (fun s -> other_set.(s) <- true) eclosure.(p))
            (Auto.trans a q s.[i])
      done;
      loop other_set (i + 1) qset)
  in
  let qset = mkset () in
  let other_set = mkset () in
  List.iter (fun q -> qset.(q) <- true) eclosure.(0);
  loop qset 0 other_set

(* Complétion d'un automate. L'alphabet complet est donné,
   un unique état puis est rajouté
*)
let is_complete auto sigma =
  let sigma = List.sort_uniq Char.compare sigma in
  try
    for q = 0 to Auto.size auto - 1 do
      let sym_trans =
        List.fold_left
          (fun acc (ch, _) -> match ch with Some c -> c :: acc | None -> acc)
          []
          (Auto.all_trans_opt auto q)
      in
      let ctrs = List.sort_uniq Char.compare sym_trans in
      if sigma <> ctrs then raise Exit
    done;
    true
  with _ -> false

let rec remove e l =
  match l with
  | [] -> []
  | x :: ll -> if x < e then x :: remove e ll else if x = e then ll else l

let complete auto sigma =
  let sigma = List.sort_uniq Char.compare sigma in
  if is_complete auto sigma then Auto.copy auto
  else
    let nsize = Auto.size auto + 1 in
    let nauto = Auto.create nsize in
    let bot = nsize - 1 in
    List.iter (fun c -> Auto.add_trans nauto bot c bot) sigma;
    for q = 0 to nsize - 2 do
      if Auto.is_final auto q then Auto.set_final nauto q;
      let to_add = ref sigma in
      let trans = Auto.all_trans_opt auto q in
      List.iter
        (fun (ch, p) ->
          let () =
            match ch with Some c -> to_add := remove c !to_add | None -> ()
          in
          Auto.add_trans_opt nauto q ch p)
        trans;
      List.iter (fun c -> Auto.add_trans nauto q c bot) !to_add
    done;
    nauto

(* renvoie une copie de l'automate avec certains sommets supprimés,
   certains symboles qui n'apparaissent que sur ces sommets peuvent
   aussi être supprimés.
   Si l'état 0 est supprimé ou si tous les états acceptants sont supprimés,
   l'automate réduit à un état sans transition est renvoyé.
*)

(* List des états non-accessible *)
let array_elements a =
  let t_elem = ref [] and f_elem = ref [] in
  for q = Array.length a - 1 downto 0 do
    if a.(q) then t_elem := q :: !t_elem else f_elem := q :: !f_elem
  done;
  (!t_elem, !f_elem)

let forward auto =
  let visited = Array.make (Auto.size auto) false in
  let rec loop q =
    if not visited.(q) then (
      visited.(q) <- true;
      let trs = Auto.all_trans_opt auto q in
      List.iter (fun (_, s) -> loop s) trs)
  in
  loop 0;
  array_elements visited

let backward auto =
  let n = Auto.size auto in
  let rev_auto = Array.make n [] in
  let set_pred q (_, p) = rev_auto.(p) <- q :: rev_auto.(p) in
  let roots = ref [] in
  for q = 0 to n - 1 do
    if Auto.is_final auto q then roots := q :: !roots;
    let trs = Auto.all_trans_opt auto q in
    List.iter (set_pred q) trs
  done;
  let visited = Array.make (Auto.size auto) false in
  let rec loop q =
    if not visited.(q) then (
      visited.(q) <- true;
      List.iter loop rev_auto.(q))
  in
  List.iter loop !roots;
  array_elements visited

let is_empty auto =
  let f, _ = forward auto in
  not (List.exists (Auto.is_final auto) f)

(* Renvoie une copie émondée de l'automate *)
let clean auto =
  let _, nf_states = forward auto in
  let _, nb_states = backward auto in
  Auto.remove_states auto (nf_states @ nb_states)

module S = Set.Make (struct
  type t = int

  let compare = Stdlib.compare
end)

let make_det auto =
  let eclosure = eclosure auto in
  let set_closure qlist =
    List.fold_left (fun acc q -> union eclosure.(q) acc) [] qlist
  in
  (* Ensemble d'états → (is_final, (symbol → ensemble d'états))*)
  let hauto = Hashtbl.create 16 in
  let id = ref ~-1 in
  let todo = Stack.create () in
  Stack.push todo eclosure.(0);
  while not (Stack.is_empty todo) do
    let qset = Stack.pop todo in
    let htrans =
      try snd (Hashtbl.find hauto qset)
      with Not_found ->
        let h = Hashtbl.create 16 in
        Hashtbl.add hauto qset
          (incr id;
           (!id, h));
        h
    in
    List.iter
      (fun q ->
        let trs, _ = Auto.all_trans auto q in
        List.iter
          (fun (ch, succs) ->
            let old_succs = try Hashtbl.find htrans ch with Not_found -> [] in
            Hashtbl.replace htrans ch (union old_succs (set_closure succs)))
          trs)
      qset;
    Hashtbl.iter
      (fun _ qset -> if not (Hashtbl.mem hauto qset) then Stack.push todo qset)
      htrans
  done;
  let nauto = Auto.create (Hashtbl.length hauto) in
  let get_id set = fst (Hashtbl.find hauto set) in
  Hashtbl.iter
    (fun qset (q, htrans) ->
      if List.exists (Auto.is_final auto) qset then Auto.set_final nauto q;
      Hashtbl.iter
        (fun ch qset -> Auto.add_trans nauto q ch (get_id qset))
        htrans)
    hauto;
  nauto

(* Déterminisation.

      Par la méthode des sous-ensembles i.e. un état de l'automate déterministe
      est un ensemble d'états de l'automate non déterministe.

      Note : On ne cherche pas à construire un automate complet. En particulier,
      on n'ajoute pas de transition vers l'état correspondant à un ensemble vide
      (cela en ferait beaucoup avec un alphabet de 256 caractères).
      Pour obtenir un automate complet, il suffit de commenter le
      test `if not (S.is_empty s')` dans le code ci-dessous.

   let make_det a =
     assert (not (Auto.has_epsilon a));
     let transf = Vector.create (Auto.size a) ([||], false) in
     (* i -> trans, final *)
     let states = Hashtbl.create 16 in
     (* S.t -> int *)
     let todo = Stack.create () in
     (* transitions à calculer *)
     let find s =
       try Hashtbl.find states s
       with Not_found ->
         let n = Vector.size transf in
         Hashtbl.add states s n;
         Vector.push transf (Array.make 256 [], S.exists (Auto.is_final a) s);
         Stack.push todo s;
         n
     in
     ignore (find (S.singleton 0));
     (* crée l'état initial + assure que c'est 0 *)
     while not (Stack.is_empty todo) do
       let s = Stack.pop todo in
       let i = Hashtbl.find states s in
       let t, _ = Vector.get transf i in
       for c = 0 to 255 do
         let add l s' = List.fold_right S.add l s' in
         let s' =
           S.fold (fun q s' -> add (Auto.trans a q (Char.chr c)) s') s S.empty
         in
         if not (S.is_empty s') then t.(c) <- [ find s' ]
       done
     done;
     let n = Vector.size transf in
     let auto = Auto.create n in
     for q = 0 to n - 1 do
       let trans, is_final = Vector.get transf q in
       if is_final then Auto.set_final auto q;
       for c = 0 to 255 do
         match trans.(c) with
         | [] -> ()
         | [ p ] -> Auto.add_trans auto q (Char.chr c) p
         | _ -> assert false (* l'automate est déterministe *)
       done
     done;
     auto
*)

(* nombre de mots reconnus de longueur k *)
let count a k =
  if not (Auto.is_det a) then invalid_arg "count";
  let n = Auto.size a in
  let m = Matrix.make n 0 in
  for i = 0 to n - 1 do
    let trans = Auto.all_trans_opt a i in
    List.iter (fun (_, j) -> m.(i).(j) <- m.(i).(j) + 1) trans
  done;
  let v = Array.init n (fun i -> Math.iverson (Auto.is_final a i)) in
  (Matrix.power_apply m k v).(0)

let rec list_first f = function
  | [] -> None
  | x :: l -> ( match f x with None -> list_first f l | Some _ as r -> r)

let rec int_first lo hi f =
  if lo >= hi then None
  else match f lo with None -> int_first (lo + 1) hi f | Some _ as r -> r

(* un mot reconnu de longueur k

   par mémoïsation ou programmation dynamique, on calcule plus généralement
   f(s,j) un mot reconnu de longueur j<=k à partir de l'état s, et la
   solution est f(0,k) *)
let some_word a k =
  let memo = Hashtbl.create 16 in
  let rec f s j =
    assert (0 <= j && j <= k);
    try Hashtbl.find memo (s, j)
    with Not_found ->
      let w =
        if j = 0 then if Auto.is_final a s then Some "" else None
        else
          let cons c = function
            | None -> None
            | Some s -> Some (String.make 1 (Char.chr c) ^ s)
          in
          int_first 0 256 (fun c ->
              list_first
                (fun s' -> cons c (f s' (j - 1)))
                (Auto.trans a s (Char.chr c)))
      in
      Hashtbl.add memo (s, j) w;
      w
  in
  match f 0 k with None -> raise Not_found | Some s -> s

let closing c =
  match c with
  | '[' -> ']'
  | '(' -> ')'
  | '{' -> '}'
  | _ -> failwith "invalid char"

let dyck s =
  let n = String.length s in
  let rec loop stack i =
    if i = n then stack = []
    else
      match (s.[i], stack) with
      | (('[' | '(' | '{') as c), _ -> loop (c :: stack) (i + 1)
      | (']' | ')' | '}'), [] -> false
      | ((']' | ')' | '}') as c), p :: sstack ->
          c = closing p && loop sstack (i + 1)
      | _ -> false
  in
  loop [] 0
