module Stack = Istack (* nos piles *)

(** acceptation de la chaine [s] par l'automate [a] avec backtracking *)
let accept a s =
  let n = String.length s in
  let rec loop i q =
    if i = n then Auto.is_final a q
    else
      exists (loop i) (Auto.eps_trans a q)
      || exists (loop (i + 1)) (Auto.trans a q s.[i])

  and exists f l = (* On inline la définition de List.exist
                      pour avoir un appel terminal dans le cas
                      d'automates déterministes *)
    match l with
      [] -> false
    | [ q ] -> f q  (* appel terminal *)
    | q :: ll -> f q || exists f ll
  in
  loop 0 0

(** Formulation impérative *)
let accept_imp a s =
  let todo = Stack.create () in
  let res = ref false in
  let n = String.length s in
  Stack.push todo (0, 0);
  while not (Stack.is_empty todo) do
    let q, i = Stack.pop todo in
    if i = n && Auto.is_final a q then res := true
    else (
      List.iter (fun q2 -> Stack.push todo (q2, i)) (Auto.eps_trans a q);
      List.iter (fun q2 -> Stack.push todo (q2, i + 1)) (Auto.trans a q s.[i]))
  done;
  !res

(** De nombreuses opérations sont basées sur un parcours du graphe de l'automate.
    On écrit une fonction générique de DFS sur un automate, puis on exprime
    les opérations en fonction de cette dernière.
*)

(** Fonctions utilitaires *)

(** [array_elements a] prend en argument un tableau de booléens et
    renvoie une paire de liste des indices valant true et des indices
    valant false *)
 let array_elements a =
  let t_elem = ref [] and f_elem = ref [] in
  for q = Array.length a - 1 downto 0 do
    if a.(q) then t_elem := q :: !t_elem else f_elem := q :: !f_elem
  done;
  (!t_elem, !f_elem)

(** Union de deux listes triées *)
let rec union_list l1 l2 =
  match (l1, l2) with
    | [], _ -> l2
    | _, [] -> l1
    | p1 :: ll1, p2 :: ll2 ->
        if p1 < p2 then p1 :: union_list ll1 l2
        else if p1 = p2 then p1 :: union_list ll1 ll2
        else p2 :: union_list l1 ll2

(** Suppression d'un élément dans une liste triée *)
let rec remove e l =
  match l with
    | [] -> []
    | x :: ll ->
      if x < e then
        x :: remove e ll
      else
        if x = e then ll else l

(** [exists_i f a] teste si un indice i est tel que 
    f i a.(i) est vrai. *)
let exists_i f a = Math.exists 0 (Array.length a) (fun i -> f i a.(i))

(** [nfa_accept a s] : acceptation par un automate non déterministe sans
    retour sur trace. *)
let nfa_accept a s =
  assert (not (Auto.has_epsilon a));
  let n = String.length s in
  let na = Auto.size a in
  let rec loop qset i other_set =
    if i = n then exists_i (fun q b -> b && Auto.is_final a q) qset
    else
      let found = ref false in
      Array.fill other_set 0 na false;
      for q = 0 to na - 1 do
        if qset.(q) then
          List.iter
            (fun p ->
              found := true;
              other_set.(p) <- true)
            (Auto.trans a q s.[i])
      done;
      !found && loop other_set (i + 1) qset
  in
  let qset = Array.make na false in
  let other_set = Array.make na false in
  qset.(0) <- true;
  loop qset 0 other_set



(** [iter_dfs succ post init] effectue un parcours en profondeur
    d'abord, à partir du sommet init.
    À chaque nouveau sommet [s], [succ s] est appelé pour connaître
    la liste des voisins de [s] (et doit renvoyer la liste vide pour
    arrêtre le parcours).
    Après parcours d'un sommet et de ses voisins, [post s (succ s)] est
    appelé pour effectuer un traitement.
*)
let iter_dfs succ post init =
  let rec loop elem =
    List.iter
    (fun elem2 ->
      loop elem2;
      post elem elem2)
      (succ elem)
    in
    loop init

(** [forward auto] renvoie une paire de listes [(la, ln)], où
    [la] est la liste des états accessibles depuis l'état initial
    et [ln] est la liste des états non accessibles.
*)
let forward auto =
  let visited = Array.make (Auto.size auto) false in
  let succ q =
    if visited.(q) then [] (* on est déjà passé par l'état *)
    else (  (* sinon on marque comme visité
               et on renvoie les successeurs, i.e. les
               sommets q' tels que q -a-> q' ou
               q -ε-> q' dans l'automate. *)
      visited.(q) <- true;
      List.map snd (Auto.all_trans_opt auto q))
  in
  iter_dfs succ (fun _ _ -> ()) 0;
  array_elements visited

(** [backward auto] renvoie une paire de listes [(lc, ln)], où
    [lc] est la liste des états co-accessibles depuis l'état initial
    et [ln] est la liste des états non co-accessibles.
 *)
let backward auto =
  (* On effectue un preprocessing de l'automate pour inverser
     les transitions. rev_auto.(q) contient la liste des prédécesseurs
     de q *)
  let n = Auto.size auto in
  let rev_auto = Array.make n [] in
  let set_pred q (_, p) = rev_auto.(p) <- q :: rev_auto.(p) in
  (* roots contient la liste des états acceptants *)
  let roots = ref [] in
  for q = 0 to n - 1 do
    if Auto.is_final auto q then roots := q :: !roots;
    let trs = Auto.all_trans_opt auto q in
    List.iter (set_pred q) trs
  done;
  (* Comme pour forward *)
  let visited = Array.make (Auto.size auto) false in
  let succ q =
    if visited.(q) then []
    else (
      visited.(q) <- true;
      rev_auto.(q))
  in
  (* On commence le DFS depuis tous les états acceptants *)
  List.iter (iter_dfs succ (fun _ _ -> ())) !roots;
  array_elements visited


(** [is_empty auto] renvoie true si et seulement si
    l'automate reconnaît le langage vide.  *)
let is_empty auto =
  let f, _ = forward auto in
  not (List.exists (Auto.is_final auto) f)

(** [clean auto] renvoie une copie émondée de l'automate *)
let clean auto =
  let _, nf_states = forward auto in
  let _, nb_states = backward auto in
  Auto.remove_states auto (nf_states @ nb_states)


(** [is_somplete auto sigma] teste si l'automate est complet par
    rapport à un alphabet sigma donné.*)
let is_complete auto sigma =
    let sigma = List.sort_uniq Char.compare sigma in
    try
      for q = 0 to Auto.size auto - 1 do
        let sym_trans =
          List.fold_left
            (fun acc (ch, _) -> match ch with Some c -> c :: acc | None -> acc)
            []
            (Auto.all_trans_opt auto q)
        in
        let ctrs = List.sort_uniq Char.compare sym_trans in
        if sigma <> ctrs then raise Exit (* utilisation d'une exception
           pour simuler un 'break' et sortir de la boucle for en avance.
          *)
      done;
      true
    with _ -> false

(** [complete auto sigma] : renvoie un automate complet par rapport à un alphabet
    sigma.
*)
let complete sigma auto =
  let sigma = List.sort_uniq Char.compare sigma in
  if is_complete auto sigma then
    (* si déjà complet on fait juste une copie *)
    Auto.copy auto
  else
    let nsize = Auto.size auto + 1 in (* ajout d'un état puis *)
    let nauto = Auto.create nsize in
    let bot = nsize - 1 in
    List.iter (fun c -> Auto.add_trans nauto bot c bot) sigma;
    for q = 0 to nsize - 2 do
      if Auto.is_final auto q then Auto.set_final nauto q;
      let to_add = ref sigma in
      let trans = Auto.all_trans_opt auto q in
      List.iter
        (fun (ch, p) ->
          let () =
            (* on retire de simga les symboles pour lesquesl il y a au moins
               une transition. *)
            match ch with Some c -> to_add := remove c !to_add | None -> ()
          in
          Auto.add_trans_opt nauto q ch p)
        trans;
      (* tout symbole encore présent va dans le puis *)
      List.iter (fun c -> Auto.add_trans nauto q c bot) !to_add
    done;
    nauto


(** [eclosure auto] renvoie un tableau [tab] tel que
    tab.(q) est l'ε-fermeture de q. *)
let eclosure auto =
  (* plutot que d'utiliser un tableau de booléens, on
     utilise le tableau de liste directement. Si visited.(q)
     contient la liste vide ce dernier n'est pas visité,
     sinon il contient au moins q. *)
  let visited = Array.make (Auto.size auto) [] in
  let succ q =
    match visited.(q) with
    | [] ->
        visited.(q) <- [ q ];
        Auto.eps_trans auto q (* on ne considère que les  ε-transitions*)
    | _ -> []
  in
  let post q p = visited.(q) <- union_list visited.(q) visited.(p) in
  iter_dfs succ post 0;
  visited

(** [product comb auto1 auto2] calcule le produit d'automate des deux automates.
    (q1, q2) devient un état acceptant dans le produit si [comb q1 q2] vaut vrai.
    En plus de la construction théorique, on doit s'occuper de certains détails
    pratiques :
    - il nous faudra reconstruire un automate dont avec un état initial 0,
    et des états qui sont des entiers de 0 à n-1 (et non pas des paires d'entiers).
    À chaque fois qu'on génère une nouvelle paire (q1, q2), on lui associe un nouvel
    entier unique qui sera l'état correspondant dans l'automate final.
    - on doit garder la trace des transitions entre (q1,q2) et tous les (q'1, q'2).
    On utilise donc une table de hachage dont les clés sont les paires d'états
    créés.
    Une fois que toutes ces structures sont en place, on peut faire un simple dfs.
*)
let product comb auto1 auto2 =
    (* les transitions de l'automate résultat *)
    let trans = Hashtbl.create 16 in
    (* fonction auxiliaire qui crée des entiers unique *)
    let new_id =
      let c = ref ~-1 in
        fun () -> incr c; !c
    in
    (* fonction utilitaire qui ajoute une transition p -c-> q dans
        l'automate résultat. Si p n'est pas encore connu, on lui
        associe un identifiant unique.*)
    let add_trans p c q =
      let id, lst =
        try Hashtbl.find trans p with Not_found -> (new_id (), [])
      in
      Hashtbl.replace trans p (id, (c, q) :: lst)
    in
    (* Calcul du successeur pour notre DFS*)
    let succ q =
      let q1, q2 = q in (* un état résultat
                            est une paire d'état de auto1 et auto2*)
      if Hashtbl.mem trans q then []
      else
        (* double boucle sur les transitions sortantes de q1 et q2
        *)
        let trs1 = Auto.all_trans_opt auto1 q1 in
        let trs2 = Auto.all_trans_opt auto2 q2 in
        List.fold_left
          (fun acc (c1, p1) ->
            List.fold_left
              (fun acc (c2, p2) ->
                if c1 = c2 then (
                  add_trans q c1 (p1, p2);
                  (p1, p2) :: acc)
                else acc)
              acc trs2)
          [] trs1
    in
    (* on commence par dire que l'état initial est la paire des
        deux états initiaux. *)
    Hashtbl.add trans (0, 0) (new_id (), []);
    iter_dfs succ (fun _ _ -> ()) (0, 0);

    (* Après le DFS, on construit l'automate grace à notre API
        en parcourant la table de hachage. *)
    let nauto = Auto.create (Hashtbl.length trans) in
    Hashtbl.iter
      (fun (q1, q2) (id, trs) ->
        if comb (Auto.is_final auto1 q1) (Auto.is_final auto2 q2) then
          Auto.set_final nauto id;
        List.iter
          (fun (c, p) ->
            Auto.add_trans_opt nauto id c (fst (Hashtbl.find trans p)))
          trs)
      trans;
    nauto

let inter auto1 auto2 = product ( && ) auto1 auto2

(** [rem_epsilon auto] supprime les transitions spontannées,
    en laissant l'automate non déterministe.*)
let rem_epsilon auto =
  if not (Auto.has_epsilon auto) then Auto.copy auto
  else
    let ecl = eclosure auto in
    let ltrs = ref [] in
    let nauto = Auto.create (Auto.size auto) in
    for q = 0 to Auto.size auto - 1 do
      for ic = 0 to 255 do
        let c = Char.chr ic in
        List.iter (fun p ->
          let trs = Auto.trans auto p c in
          List.iter (fun s -> ltrs := (q,c,s) :: !ltrs) trs
          ) ecl.(q)
      done;
      if Auto.is_final auto q then Auto.set_final nauto q;
    done;
    List.iter (fun (q,c,p) -> Auto.add_trans nauto q c p) !ltrs;
    nauto

(** [enfa_union] renvoie l'union de deux automates comme
    un automate avec transitions spontannées. *)
let enfa_union auto1 auto2 =
  let size1 = Auto.size auto1 in
  let size2 = Auto.size auto2 in
  let nsize = size1 + size2 + 1 in
  let nauto = Auto.create nsize in
  (* On ajoute les anciennes transitions en renommant *)
  let todo = [ auto1, size1, 1; auto2, size2, size1 + 1 ] in
  List.iter (fun (auto, size, offset) ->
    for q = 0 to size - 1 do
      List.iter (fun (c, p) ->
        Auto.add_trans_opt nauto (q + offset) c (p + offset);
        if Auto.is_final auto q then
          Auto.set_final nauto (q + offset)
        ) (Auto.all_trans_opt auto q)
    done
    ) todo;
  Auto.add_eps_trans nauto 0 1;
  Auto.add_eps_trans nauto 0 (size1 + 1);
  nauto

(** [sigma auto] renvoie tous les caractères présents explicitement
    dans l'automate données. *)
let sigma auto =
  let chars = Array.make 256 false in
  for q = 0 to Auto.size auto - 1 do
    let trs, _ = Auto.all_trans auto q in
    List.iter (fun (c, _) -> chars.(Char.code c) <- true) trs;
  done;
  List.map Char.chr (fst (array_elements chars))


(** [union auto1 auto2] renvoie l'automate union en utilisant le produit
    d'automates. On doit par contre avoir des automates complets.
    On préserve le caractère de l'automate :
    - union entre enfa donne un enfa
    - union entre nfa donne un nfa
    - union entre dfa donne un dfa
    *)
let union auto1 auto2 =
  if Auto.has_epsilon auto1 || Auto.has_epsilon auto2 then
    enfa_union auto1 auto2
  else
    let sigma1 = sigma auto1 in
    let sigma2 = sigma auto2 in
    let nsigma = union_list sigma1 sigma2 in
    let nauto1 = complete nsigma auto1 in
    let nauto2 = complete nsigma auto2 in
    product ( || ) nauto1 nauto2

(** [make_det auto] déterminise l'automate grace à la construction par sous
    ensemble. *)

let make_det auto =
  if Auto.is_det auto then Auto.copy auto
  else
    let auto = if Auto.has_epsilon auto then rem_epsilon auto else auto in
    let all_chars = sigma auto in
    (* même structure que pour l'automate produit *)
    (* les transitions de l'automate résultat *)
    let trans = Hashtbl.create 16 in
    (* fonction auxiliaire qui crée des entiers unique *)
    let new_id =
      let c = ref ~-1 in
        fun () -> incr c; !c
    in
    (* fonction utilitaire qui ajoute une transition p -c-> q dans
        l'automate résultat. Si p n'est pas encore connu, on lui
        associe un identifiant unique.*)
    let add_trans p c q =
      let id, lst =
        try Hashtbl.find trans p with Not_found -> (new_id (), [])
      in
      Hashtbl.replace trans p (id, (c, q) :: lst)
    in
    (* Calcul du successeur pour notre DFS*)
    let succ qset =
      if Hashtbl.mem trans qset then []
      else
        let () = Hashtbl.add trans qset (new_id (), []) in
        (* Pour chaque symbole et chaque état dans qset *)
        List.fold_left (fun acc c ->
          let succs =
            List.fold_left (fun acc q ->
              union_list acc (Auto.trans auto q c)
              ) [] qset
          in
          if succs <> [] then (
            add_trans qset c succs; succs :: acc)
          else acc
          ) [] all_chars
    in
    iter_dfs succ (fun _ _ -> ()) [0];
    (* Après le DFS, on construit l'automate grace à notre API
        en parcourant la table de hachage. *)
    let nauto = Auto.create (Hashtbl.length trans) in
    Hashtbl.iter
        (fun qset (id, trs) ->
          if List.exists (Auto.is_final auto) qset then
            Auto.set_final nauto id;
          List.iter
              (fun (c, p) ->
                Auto.add_trans nauto id c (fst (Hashtbl.find trans p)))
              trs)
          trans;
    nauto


(** [compl_dfa auto sigma] renvoie le complémentaire d'un DFA par rapport à un
    alphabet sigma *)

let compl_dfa sigma auto =
  assert (Auto.is_det auto);
  let nauto = complete sigma auto in
  for q = 0 to Auto.size nauto - 1 do
    if Auto.is_final nauto q then
      Auto.unset_final nauto q
    else
      Auto.set_final nauto q
  done;
  nauto


(** [diff sigma auto1 auto2] calcule la différence 
    de langage entre auto1 et auto2 *)
let diff sigma auto1 auto2 =
  let comp2 = compl_dfa sigma (make_det auto2) in
  inter auto1 comp2
(** [equiv sigma auto1 auto2] renvoie vrai si les automates
    reconnaissent le même langage sur sigma* *)
let equiv sigma auto1 auto2 = is_empty (diff sigma auto1 auto2)

