
(* Arbres binaires *)

type 'a bintree =
  | E
  | N of 'a bintree * 'a * 'a bintree

let rec size (t: 'a bintree) : int =
   match t with
  | E           -> 0
  | N (l, _, r) -> 1 + size l + size r

let rec perfect (h: int) : int bintree =
  if h = -1 then
    E
  else
    N (perfect (h-1), h, perfect (h-1))

let rec height t = match t with
  | E           -> -1
  | N (l, _, r) -> 1 + max (height l) (height r)

let rec preorder t = match t with
  | E -> ()
  | N (l, x, r) ->
      print_string x;
      preorder l;
      preorder r

let rec inorder t = match t with
  | E -> ()
  | N (l, x, r) ->
      inorder l;
      print_string x;
      inorder r

let rec postorder t = match t with
  | E -> ()
  | N (l, x, r) ->
      postorder l;
      postorder r;
      print_string x

module Stack = Istack (* nos piles *)

let preorder_iterative t =
  let st = Stack.create () in
  Stack.push st t;
  while not (Stack.is_empty st) do
    match Stack.pop st with
    | E -> ()
    | N (l, x, r) -> print_string x; Stack.push st r; Stack.push st l
  done

let inorder_iterative t =
  let st = Stack.create () in
  Stack.push st t;
  while not (Stack.is_empty st) do
    match Stack.pop st with
    | E -> ()
    | N (E, x, E) -> print_string x
    | N (l, x, r) -> Stack.push st r; Stack.push st (N(E,x,E)); Stack.push st l
  done

let postorder_iterative t =
  let st = Stack.create () in
  Stack.push st t;
  while not (Stack.is_empty st) do
    match Stack.pop st with
    | E -> ()
    | N (E, x, E) -> print_string x
    | N (l, x, r) -> Stack.push st (N(E,x,E)); Stack.push st r; Stack.push st l
  done

let rec left t n = if n = 0 then t else left (N (t, n, E)) (n - 1)
let left n = left E n

let rec right t n = if n = 0 then t else right (N (E, n, t)) (n - 1)
let right n = right E n

let rec random n =
  if n = 0 then
    E
  else
    let k = Random.int n in
    N (random k,
       String.make 1 (Char.chr (Char.code 'a' + Random.int 26)),
       random (n - 1 - k))

let random_equi n =
  let c = Math.catalan (n + 1) in
  let rec build n =
    if n = 0 then
      E
    else
      let i = Random.int c.(n) in
      let rec find s k =
        let s = s + c.(k) * c.(n-1-k) in
        if i < s then N (build k, string_of_int n, build (n - 1 - k))
        else find s (k + 1) in
      find 0 0
  in
  build n

(* note : plus simple en passant par un tableau avec Array.of_list *)
let of_list lst =
  let rec build n lst =
    if n = 0 then
      E, lst
    else
      let k = (n - 1) / 2 in
      let l, lst = build k lst in
      match lst with
      | [] -> assert false
      | x :: lst ->
          let r, lst = build (n - 1 - k) lst in
          N (l, x, r), lst
  in
  let t, lst = build (List.length lst) lst in
  assert (lst = []);
  t

let rec of_sub_array a lo hi =
  assert (0 <= lo && lo <= hi && hi <= Array.length a);
  if lo = hi then
    E
  else
    let mid = lo + (hi - lo) / 2 in
    N (of_sub_array a lo mid, a.(mid), of_sub_array a (mid+1) hi)

let of_array a =
  of_sub_array a 0 (Array.length a)

let is_bst t =
  let ole lo x = match lo with None -> true | Some lo -> lo <= x in
  let leo x hi = match hi with None -> true | Some hi -> x <= hi in
  let rec check lo t hi = match t with
    | E -> true
    | N (l, x, r) ->
        check lo l (Some x) && ole lo x && leo x hi && check (Some x) r hi in
  check None t None

(* tous les arbres binaires, par programmation dynamique *)
let all n =
  let trees = Array.make (n + 1) [] in
  trees.(0) <- [E];
  for i = 1 to n do
    for j = 0 to i - 1 do
      List.iter (fun l ->
        List.iter (fun r ->
          trees.(i) <- N (l, i, r) :: trees.(i))
        trees.(i - 1 - j))
      trees.(j)
    done
  done;
  trees.(n)

(* est-ce plus lisible écrit ainsi ? *)
let all n =
  let trees = Array.make (n + 1) [] in
  trees.(0) <- [E];
  for s = 1 to n do
    for sl = 0 to s - 1 do
      trees.(sl) |> List.iter @@ fun l ->
        trees.(s - 1 - sl) |> List.iter @@ fun r ->
          trees.(s) <- N (l, s, r) :: trees.(s)
    done
  done;
  trees.(n)

let card t = match t with
  | E                -> 0
  | N (_, (_, n), _) -> n

let rec count t = match t with
  | E ->
      E
  | N (l, x, r) ->
      let l = count l in
      let r = count r in
      N (l, (x, 1 + card l + card r), r)

let rec nth t i = match t with
  | E ->
      invalid_arg "nth"
  | N (l, (x, _), r) ->
      if i = card l then
        x
      else if i < card l then
        nth l i
      else
        nth r (i - card l - 1)

let deepest t =
  let rec find d t = match t with
    | N (E, x, E) -> d, x
    | N (E, _, t) | N (t, _, E) -> find (d+1) t
    | N (l, _, r) ->
        let dl, _ as xl = find (d+1) l in
        let dr, _ as xr = find (d+1) r in
        if dl > dr then xl else xr
    | E -> invalid_arg "deepest"
  in
  snd (find 0 t)


let rec maxsum t = match t with
  | E           -> 0
  | N (l, x, r) -> x + max (maxsum l) (maxsum r)

let rec maxsum_path t = match t with
  | E -> [], 0
  | N (l, x, r) ->
      let pl, sl = maxsum_path l in
      let pr, sr = maxsum_path r in
      if sl > sr then x :: pl, x + sl else x :: pr, x + sr

let rec print = function
  | E ->
      ()
  | N (l, s, r) ->
      print_char '('; print l; print_string s; print r; print_char ')'

