
(* Files de priorités réalisées avec des arbres de Braun

   L'ordre utilisé ici est la comparaison structurelle polymorphe d'OCaml. *)

type 'a heap =
  | E
  | N of 'a heap * 'a * 'a heap
    (* invariant : size(left) = size(right)
                ou size(left) = size(right) + 1 *)

let empty =
  E

let is_empty t =
  t = E

let rec insert x t = match t with
  | E ->
      N (E, x, E)
  | N (l, y, r) ->
      if x <= y then
        N (insert y r, x, l)
      else
        N (insert x r, y, l)

let rec extract t = match t with
  | E ->
      assert false
  | N (E, y, r) ->
      assert (r = E);
      y, E
  | N (l, y, r) ->
      let x, l = extract l in
      x, N (r, y, l)

let is_le x t = match t with
  | E -> true
  | N (_, y, _) -> x <= y

let rec replace_min x t = match t with
  | N (l, _, r) when is_le x l && is_le x r ->
      N (l, x, r)
  | N ((N (_, lx, _) as l), _, r) when is_le lx r ->
      N (replace_min x l, lx, r) (* car lx <= x, rx *)
  | N (l, _, (N (_, rx, _) as r)) ->
      N (l, rx, replace_min x r) (* car rx <= x, lx *)
  | E | N (E, _, _) | N (_, _, E) ->
      assert false

(* fusionne deux arbres de Braun [l] et [r],
   sous l'hypothèse que [size r <= size l <= size r + 1] *)
let rec merge l r = match l, r with
  | _, E ->
      l
  | N (ll, lx, lr), N (_, ly, _) ->
      if lx <= ly then
        N (r, lx, merge ll lr)
      else
        let x, l = extract l in
        N (replace_min x r, ly, l)
  | E, _ ->
      assert false (* contradiction *)

let get_min t = match t with
  | E -> invalid_arg "get_min"
  | N (_, x, _) -> x

let extract_min t = match t with
  | E -> invalid_arg "extract_min"
  | N (l, x, r) -> x, merge l r

(* calcul de la taille en temps log(n)^2
   (bien entendu, on pourrait encore plus simplement
    maintenir la taille à côté de l'arbre) *)

(* si `m <= size t <= m+1` alors `diff m t = size t - m` *)
let rec diff m = function
  | E -> 0
  | N _ when m = 0 -> 1
  | N (l, _, r) ->
      if m mod 2 = 1 then
        diff (m / 2) l
      else
        diff ((m - 1) / 2) r

let rec size = function
  | E -> 0
  | N (l, _, r) ->
      let m = size r in 1 + 2 * m + diff m l

let rec iter f = function
  | E -> ()
  | N (l, x, r) -> f x; iter f l; iter f r

let rec fold f acc = function
  | E ->
      acc
  | N (l, x, r) ->
      let acc = f acc x in
      let acc = fold f acc l in
      fold f acc r

