
(* Programmation dynamique

   Somme maximale dans une pyramide d'entiers

   Une pyramide d'entiers est donnée sous la forme d'un tableau
   de n tableaux.
   Les éléments sont aux indices (i,j) avec 0 <= j <= i < n.
   Voici un exemple avec n = 5 :

     [|   [| 3 |];
         [| 1;4|];
        [| 1;5;9|];
       [| 2;6;5;3|];
      [| 5;8;9;7;9|]; |]

   On cherche à maximiser la somme en parcourant la pyramide en haut
   en bas, en n'étant seulement autorisé à descendre vers l'un des
   deux entiers situés immédiatement en dessous.

   Sur l'exemple ci-dessus, la somme maximale est 30.
*)

(* version récursive, exponentielle *)
let max_sum_rec (a: int array array) : int =
  let rec s i j = (* 0 <= j <= i <= n *)
    if i = Array.length a then 0
    else a.(i).(j) + max (s (i+1) j) (s (i+1) (j+1))
  in
  s 0 0

(* calcul par mémoïsation *)
let max_sum_memo (a: int array array) : int =
  let memo = Hashtbl.create 16 in
  let rec s i j = (* 0 <= j <= i <= n *)
    try
      Hashtbl.find memo (i, j)
    with Not_found ->
      let v =
        if i = Array.length a then 0
        else a.(i).(j) + max (s (i+1) j) (s (i+1) (j+1)) in
      Hashtbl.add memo (i, j) v;
      v
  in
  s 0 0

(* calcul de bas en haut, temps O(n^2) espace O(n^2) *)
let max_sum_dp (a: int array array) : int =
  let n = Array.length a in
  let s = Array.make_matrix (n+1) (n+1) 0 in (* 0 <= j <= i <= n *)
  for i = n - 1 downto 0 do
    for j = 0 to i do
      s.(i).(j) <- a.(i).(j) + max s.(i+1).(j) s.(i+1).(j+1)
    done
  done;
  s.(0).(0)

(* programmation dynamique, temps O(n^2) espace O(n) *)
let max_sum_dp2 (a: int array array) : int =
  let n = Array.length a in
  let s = Array.make (n+1) 0 in
  for i = n - 1 downto 0 do
    for j = 0 to i do
      s.(j) <- a.(i).(j) + max s.(j) s.(j+1)
    done
  done;
  s.(0)

(* avec la solution, c'est-à-dire les indices utilisés *)
let max_sum_sol (a: int array array) : int * (int * int) list =
  let n = Array.length a in
  let s = Array.make (n+1) 0 in (* 0 <= i <= n *)
  let sol = Array.make (n+1) [] in
  for i = n - 1 downto 0 do
    for j = 0 to i do
      sol.(j) <- (i, j) :: if s.(j) > s.(j+1) then sol.(j) else sol.(j+1);
      s.(j) <- a.(i).(j) + max s.(j) s.(j+1)
    done
  done;
  (s.(0), sol.(0))

