
(* Calcul propositionnel

   Les variables sont ici des entiers entre 1 et n.
   Cela permet notamment une représentation facile des littéraux
   (entier i ou -i) et une forme normale (CNF ou DNF) qui est
   directement au format DIMACS.
*)

open Format

type binop = And | Or | Imp

type fmla =
  | True
  | False
  | Var of int (* dans 1..n *)
  | Not of fmla
  | Bin of binop * fmla * fmla

let string_of_binop = function
  | And     -> "/\\"
  | Or      -> "\\/"
  | Imp -> "->"

let rec print fmt = function
  | True ->
      fprintf fmt "true"
  | False ->
      fprintf fmt "false"
  | Var i ->
      fprintf fmt "%d" i
  | Not f ->
      fprintf fmt "(@[not@ %a@])" print f
  | Bin (op, f1, f2) ->
      fprintf fmt "(@[%a %s@ %a@])" print f1 (string_of_binop op) print f2

let rec size = function
  | True | False | Var _ -> 1
  | Not f                -> 1 + size f
  | Bin (_, f1, f2)      -> 1 + size f1 + size f2

let rec subst i ps f = match f with
  | Var j when i = j -> ps
  | Var _            -> f
  | True | False     -> f
  | Not f            -> Not (subst i ps f)
  | Bin (op, f1, f2) -> Bin (op, subst i ps f1, subst i ps f2)

(* Une valuation associe une valeur de vérité à chaque variable.
   La case 0 du tableau est inutilisée. *)

type valuation = bool array

let rec eval v = function
  | True  -> true
  | False -> false
  | Var i -> assert (1 <= i && i < Array.length v); v.(i)
  | Not f -> not (eval v f)
  | Bin (And, f1, f2) -> eval v f1 && eval v f2
  | Bin (Or,  f1, f2) -> eval v f1 || eval v f2
  | Bin (Imp, f1, f2) -> not (eval v f1) || eval v f2

(* Pour manipuler des tables de vérité ou des valuations, il peut être
   utile de commencer par déterminer l'ensemble des variables. *)

let varmax f =
  let rec varmax m = function
    | True | False    -> m
    | Var i           -> max i m
    | Not f           -> varmax m f
    | Bin (_, f1, f2) -> varmax (varmax m f1) f2 in
  varmax 0 f

(* Table de vérité

   Affiche les 2^n lignes de la table, dans l'ordre lexicographique
   donné par F < T. *)

let print_truth_table fmt f =
  let nv = varmax f in
  let v = Array.make (nv + 1) false in
  let bool fmt b = fprintf fmt "%s" (if b then "T" else "F") in
  let rec fill i =
    if i = nv + 1 then (
      for x = 1 to nv do fprintf fmt "%a" bool v.(x) done;
      fprintf fmt " => %a@\n" bool (eval v f)
    ) else (
      v.(i) <- false; fill (i+1); v.(i) <- true;  fill (i+1)) in
  fill 1


(* Quelques raccourcis et constructions

   Note : les fonctions comme `f_or` pourraient effectuer des
   simplifications, comme par exemple simplifier
   `Bin(Or, f, True)` en `f`.
   Voir les fonctions de construction définies pour la simplification
   des formules dans l'algorithme de Quine (quine.ml) *)

let f_true = True
let f_false = False
let f_var i = Var i
let f_not f = Not f
let f_or f1 f2 = Bin (Or, f1, f2)
let f_and f1  f2 = Bin (And, f1, f2)
let f_implies f1  f2 = Bin (Imp, f1, f2)
let f_xor f1 f2 = f_or (f_and f1 (Not f2)) (f_and (Not f1) f2)
let f_equiv f1 f2 = f_and (f_implies f1 f2) (f_implies f2 f1)
