/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// Table de hachage (string -> int)

#include "prelude.h"
#include <limits.h>

#include "hash.h"
#include "hashtbl.h"

typedef struct Entry {
  char *key;
  int value;
  struct Entry *next;
} entry;

struct Hashtbl {
  int capacity; // 0 < capacity
  int size;     // nombre d'entrées dans la table
  entry **data;  // tableau de taille capacity
};

hashtbl *hashtbl_create(int capacity) {
  assert(0 < capacity);
  hashtbl *h = malloc(sizeof(struct Hashtbl));
  h->capacity = capacity;
  h->data = calloc(capacity, sizeof(entry*));
  for (int i = 0; i < capacity; i++) {
    h->data[i] = NULL;
  }
  h->size = 0;
  return h;
}

int hashtbl_size(hashtbl *h) {
  return h->size;
}

int hashtbl_bucket(hashtbl *h, char *k) {
  int i = hash(k) & INT_MAX; // s'assurer que i >= 0
  return i % h->capacity;    // avant de prendre le modulo
}
// alternative : capacity est une puissance de deux
// et on fait alors hash(k) & (h->capacity-1)
// ce qui épargne le & INT_MAX et surtout évite une division coûteuse

entry *hashtbl_find_entry(entry *e, char *k) {
  while (e != NULL) {
    if (strcmp(k, e->key) == 0) {
      return e;
    }
    e = e->next;
  }
  return NULL;
}

entry *create_entry(char *k, int v, entry *n) {
  entry *e = malloc(sizeof(struct Entry));
  e->key = k;
  e->value = v;
  e->next = n;
  return e;
}

void hashtbl_put(hashtbl *h, char *k, int v) {
  int b = hashtbl_bucket(h, k);
  entry *e = hashtbl_find_entry(h->data[b], k);
  if (e == NULL) {
    h->data[b] = create_entry(k, v, h->data[b]);
    h->size++;
  } else {
    e->value = v;
  }
}

bool hashtbl_contains(hashtbl *h, char *k) {
  int b = hashtbl_bucket(h, k);
  return hashtbl_find_entry(h->data[b], k) != NULL;
}

int hashtbl_get(hashtbl *h, char *k) {
  int b = hashtbl_bucket(h, k);
  entry *e = hashtbl_find_entry(h->data[b], k);
  if (e == NULL) return -1;
  return e->value;
}

// exercice ?
void hashtbl_remove(hashtbl *h, char *k) {
  int b = hashtbl_bucket(h, k);
  entry *e = hashtbl_find_entry(h->data[b], k);
  if (e == NULL) return;
  // on doit supprimer e de la liste h->data[b]
  entry **p = &h->data[b];
  while (*p != e) {
    p = &(*p)->next;
  }
  *p = (*p)->next;
}

void hashtbl_print(hashtbl *h) {
  printf("[");
  for (int i = 0; i < h->capacity; i++) {
    for (entry *e = h->data[i]; e != NULL; e = e->next) {
      printf("%s -> %d; ", e->key, e->value);
    }
  }
  printf("]\n");
}

void hashtbl_stat(hashtbl *h) {
  int longest = 0;
  for (int i = 0; i < h->capacity; i++) {
    int len = 0;
    for (entry *e = h->data[i]; e != NULL; e = e->next) {
      len++;
    }
    if (len > longest) longest = len;
  }
  printf("longest bucket = %d\n", longest);
}

void hashtbl_delete(hashtbl *h) {
  for (int i = 0; i < h->capacity; i++) {
    entry *e = h->data[i];
    while (e != NULL) {
      entry *p = e;
      e = e->next;
      free(p);
    }
  }
  free(h->data);
  free(h);
}
