
(* ID3

   On adopte ici le vocabulaire traditionnel dans le contexte des
   arbres de décision, en appelant "attributs" les coordonnées des
   points.

   On ne considère ici que des arbres de décisions binaires. Un noeud
   Node (i, l, r) teste l'attribut i de la donnée. À gauche (l), on
   trouve les données de valeurs <= 0 ; à droite (r) les données de
   valeurs > 0.

   Il ne serait pas très difficile de généraliser à un arbre d'arité
   variable, avec autant de branches que de valeurs possibles pour
   l'attribut, ou des plages de valeurs associées à autant de
   branches.  *)

type data = float array

type label = int

type dectree =
  | Leaf of label
  | Node of int * dectree * dectree

let rec lookup dt p =
  match dt with
  | Leaf l           -> l
  | Node (i, tl, tr) -> lookup (if p.(i) <= 0. then tl else tr) p

let rec print fmt = function
  | Leaf l ->
      Format.fprintf fmt "Leaf %d" l
  | Node (i, tl, tr) ->
      Format.fprintf fmt "@[<hov 2>Node %d@\n%a@\n%a@]" i print tl print tr

let rec size = function
  | Leaf _           -> 0
  | Node (_, tl, tr) -> 1 + size tl + size tr

let debug = true

(* on manipule ci-dessus des ensembles d'indices dans le tableau `data`
   (notés `set`) et des ensembles d'attributs (notés `att`) *)
module S = Set.Make(Int)

let interval_set lo hi = (* { lo,...,hi-1 } *)
  Math.fold_interval lo hi S.empty (fun s i -> S.add i s)

let same_class data set =
  assert (not (S.is_empty set));
  let l = snd data.(S.choose set) in
  S.for_all (fun i -> snd data.(i) = l) set

let log2 x = log x /. log 2.

(* l'attribut dans `att` qui maximise le gain, pour les données dans `set` *)
let max_gain data set att =
  let n = S.cardinal set in
  assert (n >= 2);
  let entropy d p =
    let h = Hashtbl.create 16 in (* label -> count *)
    let s = ref 0 in
    let add l =
      incr s;
      try incr (Hashtbl.find h l) with Not_found -> Hashtbl.add h l (ref 1) in
    S.iter (fun i -> let x, l = data.(i) in if p x.(d) then add l) set;
    if !s = 0 then 0. else
    Hashtbl.fold (fun _ c e ->
        if !c = 0 then e else let r = float !c /. float !s in e -. r *. log2 r)
      h 0.
  in
  let gain d =
    -. entropy d (fun x -> x <= 0.) -. entropy d (fun x -> x > 0.) in
  S.fold (fun d (dmax, gmax as acc) ->
      let g = gain d in if g > gmax then (d, g) else acc)
    att (0, neg_infinity)

(* sépare les données selon l'attribut d *)
let split data set d =
  S.fold (fun i (l, r) ->
      let x, _ = data.(i) in
      if x.(d) <= 0. then S.add i l, r else l, S.add i r)
    set (S.empty, S.empty)

let plurality data s =
  let a = Array.of_list (S.elements s) in
  Learning.plurality (Array.map (fun i -> data.(i)) a)

let build (data: (data * label) array) : dectree =
  let rec build parents set att =
    assert (not (S.is_empty parents));
    if S.is_empty set then (
      Format.eprintf "no more data@.";
      Leaf (plurality data parents)
    ) else if same_class data set then (
      Format.eprintf "same class@.";
      Leaf (snd data.(S.choose set))
    ) else if S.is_empty att then (
      Format.eprintf "no more attributes@.";
      Leaf (plurality data set)
    ) else (
      assert (S.cardinal set >= 2);
      let d, g = max_gain data set att in
      Format.eprintf "max gain: d=%d g=%.2f@." d g;
      let sl, sr = split data set d in
      let att = S.remove d att in
      Node (d, build set sl att, build set sr att)
    ) in
  let set = interval_set 0 (Array.length data) in
  let att = interval_set 0 (Array.length (fst data.(0))) in
  build set set att

let id3 data =
  if Array.length data = 0 then invalid_arg "id3";
  lookup (build data)
