
(* Arbres k-dimensionnels *)

type point = float array

type 'a kdtree =
  | Empty
  | Node of int * 'a kdtree * point * 'a * 'a kdtree
    (* dans Node(i, l, x, v, r), on a des valeurs < x dans l
       et des valeurs >= x dans r, pour la dimension i *)

let empty =
  Empty

let compare i x y =
  Stdlib.compare x.(i) y.(i)

let comparep (i: int) (x: point * 'a) (y: point * 'a) : int =
  Stdlib.compare (fst x).(i) (fst y).(i)

let rec add (i: int) (p: point) (v: 'a) (t: 'a kdtree) : 'a kdtree =
  match t with
  | Empty ->
      Node (i, Empty, p, v, Empty)
  | Node (ix, l, px, vx, r) ->
      assert (ix = i);
      let k = Array.length p in
      assert (k = Array.length px);
      let c = compare i p px in
      let i1 = (i + 1) mod k in
      if c < 0 then
        Node (i, add i1 p v l, px, vx, r)
      else
        Node (i, l, px, vx, add i1 p v r)

let add (p: point) (v: 'a) (t: 'a kdtree) : 'a kdtree =
  add 0 p v t

let rec create (i: int) (a: (point * 'a) array) (lo: int) (hi: int) : 'a kdtree =
  if lo = hi then
    Empty
  else (
    let mid = lo + (hi - lo) / 2 in
    let mid = Arrays.split (comparep i) lo mid hi a in
    let p, v = a.(mid) in
    let k = Array.length p in
    let i1 = (i + 1) mod k in
    Node (i, create i1 a lo mid, p, v, create i1 a (mid+1) hi)
  )

let of_array (pa: (point * 'a) array) : 'a kdtree =
  create 0 pa 0 (Array.length pa)

let rec iter f = function
  | Empty -> ()
  | Node (_, l, p, v, r) -> f p v; iter f l; iter f r

let array_of_pqueue pq =
  let n = Pqueue.size pq in
  if n = 0 then [||] else
  let _, qv = Pqueue.extract_min pq in
  let a = Array.make n qv in
  let i = ref 0 in
  Pqueue.iter (fun (_, qv) -> incr i; a.(!i) <- qv) pq;
  a

(* Attention : Quand on parle d'arbre k-dimensionnel ci-dessus, k
   désigne la dimension et non le nombre de points les plus proches
   dans l'arbre qui nous intéressent quand on utilise un arbre k-d
   pour l'algorithme des k plus proches voisins. Du coup on adopte
   ci-dessous `n` pour le nombre de points les plus proches. *)

let sqr x = x *. x

let sqr_dist x y =
  let n = Array.length x in
  assert (Array.length y = n);
  let d = ref 0. in
  for i = 0 to n - 1 do d := !d +. sqr (x.(i) -. y.(i)) done;
  !d

let dist x y =
  sqrt (sqr_dist x y)

(* version naïve, parcourant tout l'arbre *)
let closest_naive (n: int) (p0: point) (t: 'a kdtree) : (point * 'a) array =
  let pq = Pqueue.create () in
  let visit p v =
    Pqueue.insert pq (-. dist p p0, (p, v)); (* /!\ ordre inverse *)
    if Pqueue.size pq > n then ignore (Pqueue.extract_min pq) in
  iter visit t;
  array_of_pqueue pq

let closest (n: int) (p0: point) (t: 'a kdtree) : (point * 'a) array =
  let pq = Pqueue.create () in
  let add p v =
    Pqueue.insert pq (-. dist p p0, (p, v)); (* /!\ ordre inverse *)
    if Pqueue.size pq > n then ignore (Pqueue.extract_min pq) in
  let radius () = -. fst (Pqueue.min pq) in
  let rec visit = function
    | Empty ->
        ()
    | Node (i, l, px, vx, r) ->
        let t1, t2 = if compare i p0 px < 0 then l, r else r, l in
        visit t1;
        let c = abs_float (p0.(i) -. px.(i)) in
        if Pqueue.size pq < n || radius () >= c then (
          add px vx; visit t2
        ) in
  visit t;
  array_of_pqueue pq
