
(* Apprentissage

*)

type data = float array
type label = int

let sqr x = x *. x

let sqr_dist x y =
  let n = Array.length x in
  assert (Array.length y = n);
  let d = ref 0. in
  for i = 0 to n - 1 do d := !d +. sqr (x.(i) -. y.(i)) done;
  !d

let dist x y =
  sqrt (sqr_dist x y)

let add x y =
  let n = Array.length x in
  assert (Array.length y = n);
  for i = 0 to n - 1 do x.(i) <- x.(i) +. y.(i) done

(* Algorithme des k moyennes (apprentissage non supervisé) *)

(* Dans le code ci-dessous, `i` désigne (l'indice d')une donnée
   et `j` désigne une classe (parmi [0..k[). *)

let debug2d = false

let k_means (k: int) (data: data array) : data array * label array =
  let n = Array.length data in
  assert (n >= k);
  let size = Array.length data.(0) in
  let lab = Array.make n 0 in
  (* initialisation aléatoire *)
  let mea = Arrays.sampling k data in
  (* on boucle tant qu'il n'y a pas convergence *)
  let change = ref true in
  (* on classifie en prenant la moyenne la plus proche *)
  let classify i di =
    let m = ref 0 and md = ref (sqr_dist di mea.(0)) in
    for j = 1 to k - 1 do
      let d = sqr_dist di mea.(j) in
      if d < !md then (md := d; m := j)
    done;
    if lab.(i) <> !m then (lab.(i) <- !m; change := true)
  in
  let steps = ref 0 in
  while !change do
    change := false;
    incr steps;
    if debug2d then (
      Format.printf "  means:@.";
      Array.iter (fun p -> Format.printf "  (%.1f,%.1f)@." p.(0) p.(1)) mea
    );
    Array.iteri classify data;
    if debug2d then (
      Format.printf "  lab:@.";
      Array.iteri (fun i li -> Format.printf "  %d: %d@." i li) lab
    );
    let sum = Array.init k (fun _ -> Array.make size 0.) in
    let nb = Array.make k 0 in
    (* mettre à jour les moyennes *)
    for i = 0 to n - 1 do
      let j = lab.(i) in add sum.(j) data.(i); nb.(j) <- nb.(j) + 1
    done;
    for j = 0 to k - 1 do
      if nb.(j) > 0 then (* si aucun point, on ne change rien *)
        mea.(j) <- Array.map (fun x -> x /. float nb.(j)) sum.(j)
    done
  done;
  Format.eprintf "%d steps@." !steps;
  mea, lab

(* Algorithme des k plus proches voisins (apprentissage supervisé) *)

(* la classe la plus représentée dans `points` *)
let plurality (points: (data * label) array) : label =
  let maxc = ref (-1) in
  Array.iter (fun (_, li) -> maxc := max !maxc li) points;
  let count = Array.make (!maxc + 1) 0 in
  let add (_, j) = count.(j) <- count.(j) + 1 in
  Array.iter add points;
  let m = Array.fold_left max 0 count in
  let cand = ref [] in
  for j = 0 to !maxc do if count.(j) = m then cand := j :: !cand done;
  assert (List.length !cand > 0);
  Lists.random_element !cand

let array_of_pqueue pq =
  if Pqueue.is_empty pq then invalid_arg "array_of_pqueue";
  let a = Array.make (Pqueue.size pq) (snd (Pqueue.min pq)) in
  let i = ref (-1) in
  Pqueue.iter (fun (_,x) -> incr i; a.(!i) <- x) pq;
  a

(* Parcours linéaire des `N` éléments de `data`.  On utilise une file
   de priorité contenant les `k` points les plus proches trouvés
   jusqu'à présent.  Temps O(N log k) et espace O(k). *)
let k_nn k data d =
  if k <= 0 then invalid_arg "k_nn";
  (* on cherche les `k` points les plus proches *)
  let pq = Pqueue.create () in
  let visit ((di, _) as xi) =
    Pqueue.insert pq (-. dist d di, xi); (* ordre inverse *)
    if Pqueue.size pq > k then ignore (Pqueue.extract_min pq) in
  Array.iter visit data;
  plurality (array_of_pqueue pq)

(* En utilisant un arbre k-dimensionnel
   (attention à ne pas confondre les deux `k`) *)
let k_nn_tree k data =
  if k <= 0 then invalid_arg "k_nn";
  let t = Kdtree.of_array data in
  fun d ->
    plurality (Kdtree.closest k d t)

