(* Petites manipulations de listes *)

(* longueur d'une liste, version directe *)
let rec length (l: 'a list) : int =
  match l with
  | []     -> 0
  | _ :: t -> 1 + length t

(* longueur d'une liste, sans faire déborder la pile *)
let rec length_acc n l = match l with
  | []     -> n
  | _ :: l -> length_acc (n + 1) l

let length l =
  length_acc 0 l


(* concaténation, version directe *)
let rec append (l1: 'a list) (l2: 'a list) : 'a list =
  match l1 with
  | []       -> l2
  | x1 :: t1 -> x1 :: append t1 l2


(* renversement, version inefficace *)
let rec mauvais_rev l = match l with
  | []     -> []
  | x :: t -> append (mauvais_rev t) [x]

(* renversement, version efficace *)
let rec rev_append l1 l2 = match l1 with
  | []       -> l2
  | x1 :: t1 -> rev_append t1 (x1 :: l2)

let rev l =
  rev_append l []


(* concaténation, sans faire déborder la pile *)
let append l1 l2 =
  rev_append (rev l1) l2

(* traductions listes <-> tableaux *)
let list_of_array a =
  let rec build acc i =
    if i < 0 then acc else build (a.(i) :: acc) (i - 1)
  in
  build [] (Array.length a - 1)

let rec fill a i l = match l with
  | []     -> ()
  | x :: l -> a.(i) <- x; fill a (i + 1) l

let array_of_list l = match l with
  | [] ->
      [||]
  | x :: tl ->
      let a = Array.make (List.length l) x in
      fill a 1 tl;
      a

(* version inefficace *)
let rec bad_array_of_list l = match l with
  | []     -> [||]
  | x :: l -> Array.append [|x|] (bad_array_of_list l)


(* tri par insertion (pour petites listes) *)
let rec insert x l = match l with
  | []                 -> [x]
  | y :: _ when x <= y -> x :: l
  | y :: l'            -> y :: insert x l'

let rec insertion_sort = function
  | []     -> []
  | x :: l -> insert x (insertion_sort l)


(* fusion de listes triées *)
let rec merge l1 l2 = match l1, l2 with
  | [], l | l, [] ->
      l
  | x1 :: t1, x2 :: t2 ->
      if x1 <= x2 then
        x1 :: merge t1 l2
      else
        x2 :: merge l1 t2

(* une variante qui ne fait pas déborder la pile, pour le tri fusion *)
let rec merge_acc l l1 l2 = match l1, l2 with
  | [], l' | l', [] ->
      rev_append l l'
  | x1 :: t1, x2 :: t2 ->
      if x1 <= x2 then
        merge_acc (x1 :: l) t1 l2
      else
        merge_acc (x2 :: l) l1 t2

let rec split l1 l2 = function
  | []     -> l1, l2
  | x :: r -> split l2 (x :: l1) r

let rec mergesort l = match l with
  | [] | [_] ->
      l
  | _ ->
      let l1, l2 = split [] [] l in
      merge_acc [] (mergesort l1) (mergesort l2)


(* map, version directe *)
let rec map f l = match l with
  | [] -> []
  | x::t -> f x :: map f t

(* map, sans faire déborder la pile, en prenant l'idée de rev *)
let rev_map f l =
  let rec rev_map_append l1 l2 = match l1 with
    | []       -> l2
    | x1 :: t1 -> rev_map_append t1 (f x1 :: l2)
  in
  rev_map_append l []

let map f l = rev (rev_map f l)

(* autre manière d'écrire la précédente *)
let bilbo_map f l =
  let rec there l tmp = match l with
    | e::l -> there l (f e :: tmp)
    | []   -> back [] tmp
  and back l' tmp = match tmp with
    | e::tmp -> back (e :: l') tmp
    | []     -> l'
  in
  there l []


(* tirage aléatoire dans une liste *)
let random_element l =
  let rec scan r n = function
    | []     -> r
    | x :: l -> scan (if Random.int n = 0 then x else r) (n + 1) l
  in
  match l with
  | []     -> invalid_arg "random_element"
  | x :: l -> scan x 2 l

let rec sum f = function
  | []     -> 0
  | x :: l -> f x + sum f l


(* le plus petit entier naturel qui n'apparaît pas dans `l` *)
let mex l =
  let n = List.length l in
  let free = Array.make (n+1) true in
  List.iter (fun v -> if 0 <= v && v < n then free.(v) <- false) l;
  let rec find i = if free.(i) then i else find (i+1) in
  find 0
