/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// Table de hachage (char* -> int)

#include "prelude.h"
#include <limits.h>

#include "hash.h"
#include "lptable.h"

struct Lptable {
  int capacity; // 0 < capacity
  int size;     // nombre d'entrées dans la table
  char **keys;  // tableau de taille capacity
  int *values;  // tableau de taille capacity
};
// une case libre est identifiée par keys[i] == NULL
// invariant : 0 <= size < capacity
//             c'est-à-dire qu'il y a toujours au moins une case de libre

lptable *lptable_create(int capacity) {
  assert(0 < capacity);
  lptable *h = malloc(sizeof(struct Lptable));
  h->capacity = capacity;
  h->keys = calloc(capacity, sizeof(char*));
  h->values = calloc(capacity, sizeof(int));
  h->size = 0;
  return h;
}

int lptable_size(lptable *h) {
  return h->size;
}

int lptable_bucket(lptable *h, char *k) {
  int i = hash(k) & INT_MAX; // s'assurer que i >= 0
  return i % h->capacity;    // avant de prendre le modulo
}
// alternative : capacity est une puissance de deux
// et on fait alors hash(k) & (h->capacity-1)
// ce qui épargne le & INT_MAX et surtout évite une division coûteuse

int lptable_find_slot(lptable *h, char *k) {
  int i = lptable_bucket(h, k);
  while (true) {
    if (h->keys[i] == NULL || strcmp(k, h->keys[i]) == 0) {
      return i;
    }
    i++;
    if (i == h->capacity) i = 0;
  }
}

void lptable_put(lptable *h, char *k, int v) {
  assert(h->size < h->capacity - 1);
  int i = lptable_find_slot(h, k);
  if (h->keys[i] == NULL) h->size++;
  h->keys[i] = k;
  h->values[i] = v;
}

bool lptable_contains(lptable *h, char *k) {
  int i = lptable_find_slot(h, k);
  return h->keys[i] != NULL;
}

int lptable_get(lptable *h, char *k) {
  int i = lptable_find_slot(h, k);
  if (h->keys[i] == NULL) return -1;
  return h->values[i];
}

// on vient d'enlever j et on nettoie à droite de i
void lptable_restore(lptable *h, int j, int i) {
  i++;
  if (i >= h->capacity) i = 0;
  char *ki = h->keys[i];
  if (ki == NULL) return;
  int r = lptable_bucket(h, ki);
  if ((j < r && r <= i) || (i < j && j < r) || (r <= i && i < j)) {
    // r est (cycliquement) entre j et i => on continue
    lptable_restore(h, j, i);
  } else {
    h->keys[j] = ki;
    h->values[j] = h->values[i];
    h->keys[i] = NULL;
    lptable_restore(h, i, i);
  }
}

void lptable_remove(lptable *h, char *k) {
  int j = lptable_find_slot(h, k);
  if (h->keys[j] == NULL) return;
  h->keys[j] = NULL;
  lptable_restore(h, j, j);
}

void lptable_stat(lptable *h) {
  int longest = 0;
  for (int i = 0; i < h->capacity; i++) {
    char *k = h->keys[i];
    if (k == NULL) continue;
    int j = lptable_bucket(h, k);
    int d = j <= i ? i - j : h->capacity - j + i;
    if (d > longest) longest = d;
  }
  printf("longest distance = %d\n", longest);
}

void lptable_print(lptable *h) {
  printf("[");
  for (int i = 0; i < h->capacity; i++) {
    if (h->keys[i] == NULL) continue;
    printf("%s -> %d; ", h->keys[i], h->values[i]);
  }
  printf("]\n");
}

void lptable_delete(lptable *h) {
  free(h->keys);
  free(h->values);
  free(h);
}
