(* MAX-SAT
   - solution exacte naïve
   - solution approchée par méthode probabiliste
   - solution exacte par séparation et évaluation *)

open Cnf

(* Solution exacte naïve : exploration de toutes les valuations *)
let count_sat nf v =
  List.length (List.filter (List.exists (fun x -> x > 0 = v.(abs x))) nf.clauses)

let max_sat_exact nf =
  if nf.kind <> CNF then invalid_arg "max_sat_exact";
  let n = nf.nbvars in
  let v = Array.make (n + 1) false in
  let rec explore i =
    if i > n then
      count_sat nf v
    else
      (let a = explore (i+1) in
       v.(i) <- true;
       let b = explore (i+1) in
       v.(i) <- false;
       max a b)
  in
  explore 1

(* Solution approchée par méthode probabiliste *)

(* Valuation aléatoire *)
let random_max_sat nf =
  if nf.kind <> CNF then invalid_arg "random_max_sat";
  let n = nf.nbvars in
  Array.init (n + 1) (fun _ -> Random.bool ())

(* Version dérandomisée

   Pour une forme normale conjonctive `nf` avec c clauses, dont chaque
   clause contient au moins k littéraux, la fonction `max_ksat_approx`
   renvoie une valuation satisfiant au moins c(1 - 1/2^k) clauses.
 *)
let max_ksat_approx nf =
  if nf.kind <> CNF then invalid_arg "max_ksat_approx";
  let n = nf.nbvars in
  let v = Array.make (n + 1) false in

  (* calcul du facteur multiplicatif pour représentation en virgule fixe *)
  let max_k =
    List.fold_left max 0 (List.map List.length nf.clauses) in
  let b = 1 lsl max_k in

  for i = 1 to n do
    (* probabilité conditionnelle de satisfaction d'une clause *)
    let expect_cl cl =
      let sat_literal x = abs x <= i && x > 0 = v.(abs x) in
      if List.exists sat_literal cl then b
      else let cl' = List.filter (fun x -> abs x > i) cl in
           b - b / (1 lsl List.length cl')
    in
    (* espérance conditionnelle *)
    let expect () = List.fold_left (+) 0
                      (List.map expect_cl nf.clauses) in
    let exp_f = expect () in
    v.(i) <- true;
    let exp_t = expect () in
    if exp_f > exp_t then v.(i) <- false
  done;
  v

(* Solution exacte par séparation et évaluation *)

let empty_clauses cnf =
  List.length (List.filter ((=) []) cnf)

let opposite_unit_clauses n cnf =
  let uc = Array.make (2 * n + 1) 0 in
  List.iter (fun c -> match c with
                      | x :: [] -> uc.(n + x) <- uc.(n + x) + 1
                      | _       -> ()) cnf;
  let o = ref 0 in
  for x = 1 to n do o := !o + min uc.(n - x) uc.(n + x) done;
  !o

let lb n cnf =
  empty_clauses cnf + opposite_unit_clauses n cnf

(* propagation de la valeur d'un littéral et simplification,
   en conservant les clauses contradictoires *)
let propagate x cnf =
  let rec simplify_clause c = match c with
    | []                 -> []
    | y :: _ when y = x  -> raise Exit
    | y :: c when y = -x -> simplify_clause c
    | y :: c             -> y :: simplify_clause c in
  let fold_clause acc c =
    try simplify_clause c :: acc
    with Exit -> acc
  in
  List.fold_left fold_clause [] cnf

let maxsat nf =
  if nf.kind <> CNF then invalid_arg "maxsat_bnb";
  let n, nc = nf.nbvars, List.length nf.clauses in
  let ub = ref nc in
  let rec explore i cnf =
    if i > n then (
      let l = List.length cnf in
      if l < !ub then ub := l
    ) else
      if lb n cnf < !ub then (
        explore (i + 1) (propagate i cnf);
        explore (i + 1) (propagate (-i) cnf)
      ) (* pas de else *)
  in
  explore 1 nf.clauses;
  nc - !ub
