
(* Illustration de min-max et alpha-beta avec le jeu Othello
   Cf https://fr.wikipedia.org/wiki/Othello_(jeu)

*)

open Format
open Math

type player = B | W

let other = function B -> W | W -> B
let color = function B -> 'B' | W -> 'W'

type grid = string (* longueur 64, caractères = 'B', 'W' ou '.' *)

let get g i j = g.[8*i+j]

type state = player * grid

let player (p, _) = p

let init =
  B, "...........................WB......BW.............................."

let string_of_player = function
  | B -> "noir"
  | W -> "blanc"
let print fmt (p, g) =
  for i = 0 to 7 do
    for j = 0 to 7 do fprintf fmt "%c" (get g i j) done;
    fprintf fmt "@\n";
  done;
  fprintf fmt "  %s joue" (string_of_player p)

let () =
  printf "%a@." print init

(* Coups *)

(* let inside i =
 *   0 <= i && i < 8
 *
 * let can_play p g i j =
 *   let c = color p and o = color (other p) in
 *   get g i j = '.' &&
 *   let check d di dj =
 *     let i' = i + d*di and j' = j + d*dj in
 *     inside i' && inside j' && get g i' j' = c &&
 *     for_all 1 d (fun k -> get g (i + k*di) (j + k*dj) = o) in
 *   exists 2 8 (fun d ->
 *      check d (+1) (+1) || check d (+1) ( 0) ||
 *      check d (+1) (-1) || check d ( 0) (-1) ||
 *      check d (-1) (-1) || check d (-1) ( 0) ||
 *      check d (-1) (+1) || check d ( 0) (+1)) *)

let rec sort (i1,j1 as p1) (i2,j2 as p2) =
  if i1 <= i2 then
    if i1 = i2 then
      if j1 <= j2 then (p1, p2) else (p2, p1)
    else
      (p1, p2)
  else
    sort p2 p1

(* toutes les cases entre (i1,j1) et (i2,j2) exclus sont c *)
let line g i1 j1 i2 j2 c =
  let (i1, j1), (i2, j2) = sort (i1, j1) (i2, j2) in
  i1 = i2 && for_all (j1+1) j2 (fun j -> get g i1 j = c) ||
  j1 = j2 && for_all (i1+1) i2 (fun i -> get g i j1 = c) ||
  i2-i1 = j2-j1 && for_all 1 (i2-i1) (fun k -> get g (i1+k) (j1+k) = c) ||
  i2-i1 = j1-j2 && for_all 1 (i2-i1) (fun k -> get g (i1+k) (j1-k) = c)

let can_play p g i j =
  let c = color p and o = color (other p) in
  get g i j = '.' &&
  exists 0 8 (fun i' ->
  exists 0 8 (fun j' ->
    get g i' j' = c &&
    (abs (i - i') > 1 || abs (j - j') > 1) &&
    line g i j i' j' o))

let set m i1 j1 i2 j2 c =
  let (i1, j1), (i2, j2) = sort (i1, j1) (i2, j2) in
  if i1 = i2 then for j = j1 to j2 do m.(i1).(j) <- c done else
  if j1 = j2 then for i = i1 to i2 do m.(i).(j1) <- c done else
  if j1 < j2 then for k = 0 to i2-i1 do m.(i1 + k).(j1 + k) <- c done else
                  for k = 0 to i2-i1 do m.(i1 + k).(j1 - k) <- c done

let everywhere f =
  for i = 0 to 7 do for j = 0 to 7 do f i j done done

let play (p, g) (i, j) =
  assert (can_play p g i j);
  let m = Array.make_matrix 8 8 '.' in
  everywhere (fun i' j' -> m.(i').(j') <- get g i' j');
  m.(i).(j) <- color p;
  everywhere (fun i' j' ->
    if get g i' j' = color p && line g i j i' j' (color (other p))
    then set m i j i' j' (color p));
  (other p, String.init 64 (fun k -> m.(k / 8).(k mod 8)))

let moves_ (p, g) =
  let m = ref [] in
  for i = 0 to 7 do for j = 0 to 7 do
    if can_play p g i j then m := (i, j) :: !m
  done done;
  !m

let moves = Memo.memo (fun _ ((p, g) as s) ->
  match moves_ s with
  | [] -> let s' = (other p, g) in if moves_ s' = [] then [] else [s']
  | ml -> List.map (play s) ml
)

let outcome ((_, g) as s) =
  assert (moves s = []);
  let b = ref 0 and w = ref 0 in
  everywhere (fun i j ->
    match get g i j with 'B' -> incr b | 'W' -> incr w | _   -> ());
  if !b > !w then Some B else
  if !w > !b then Some W else
  None

let () =
  List.iter (fun s -> printf "%a@." print s) (moves init)

(* min-max and alpha-beta

   On prend ici le point de vue de l'ordinateur = joueur W
   Score =
    + 10000 si W gagne
    0 si partie nul
    - 10000 si W perd
*)

let eval p = Memo.memo (fun _ (_, g) ->
  (* on décompte les positions où on peut joueur *)
  let positions p =
    fold_interval 0 8 0 (fun acc i ->
    fold_interval 0 8 acc (fun acc j ->
        if can_play p g i j then acc+1 else acc)) in
  let e = positions p - positions (other p) in
  (* à cela on ajoute les bords (x3) et les coins (x10) *)
  let iv p i j = iverson (get g i j = color p) in
  let edges p =
    let add acc i = acc + iv p i 0 + iv p i 7 + iv p 0 i + iv p 7 i in
    fold_interval 1 7 0 add in
  let e = e + 3 * (edges p - edges (other p)) in
  let corners p = iv p 0 0 + iv p 7 7 + iv p 0 7 + iv p 7 0 in
  e + 10 * (corners p - corners (other p))
)

let minmax p = Memo.memo (fun minmax (s, d) ->
  match moves s with
  | [] ->
      (match outcome s with
       | Some p' -> if p' = p then +10000 else -10000
       | None -> 0)
  | _ when d = 0 ->
      eval p s
  | s' :: l ->
      List.fold_left (if player s = p then max else min)
        (minmax (s', d-1)) (List.map (fun s' -> minmax (s', d-1)) l)
)
let minmax p =
  let f = minmax p in
  fun d s -> f (s, d)

let alphabeta p = Memo.memo (fun alphabeta (d, a, b, s) ->
  match moves s with
  | [] ->
      (match outcome s with
       | Some p' -> if p' = p then +10000 else -10000
       | None -> 0)
  | _ when d = 0 ->
      eval p s
  | l ->
      let rec loop v a b = function
        | [] -> v
        | s' :: l ->
            let e = alphabeta (d-1, a, b, s') in
            if player s = p then (* max *)
              let v = max v e in
              if v >= b then v else loop v (max a v) b l
            else (* min *)
              let v = min v e in
              if v <= a then v else loop v a (min b v) l
      in
      loop (if player s = p then min_int else max_int) a b l
)

let alphabeta p =
  let f = alphabeta p in
  fun d s -> f (d, -10000, +10000, s)

let depth = 5
let strat = minmax W
  (* memo: 2678133 entries, 569031 hits *)
let depth = 5
let strat = alphabeta W
  (* memo: 116100 entries, 7240 hits *)
let computer s =
  let sc = strat depth s in
  printf "  score = %d@." sc;
  match moves s with
  | [] ->
      assert false
  | l  ->
      let rec choose = function
        | [] -> assert false
        | s :: l -> if strat (depth-1) s = sc then s else choose l in
      choose l

(* interface graphique *)

open Graphics

let ofs = 50 and scale = 70
let u k = ofs + scale * k
let u_ x = (x - ofs) / scale

let draw ((p, g) as s) =
  auto_synchronize false;
  clear_graph ();
  for k = 0 to 8 do
    moveto (u k) (u 0); lineto (u k) (u 8);
    moveto (u 0) (u k); lineto (u 8) (u k);
  done;
  everywhere (fun i j ->
      match get g i j with
      | 'B' | 'W' as c ->
          (if c = 'B' then fill_circle else draw_circle)
            (u j + scale/2) (u (7-i) + scale/2) (scale/2-5)
      | _   -> ());
  let msg =
    if moves s = [] then match outcome s with
                        | Some p -> sprintf "%s gagne" (string_of_player p)
                        | None   -> "match nul"
    else sprintf "%s joue" (string_of_player p) in
  moveto 200 20; draw_string msg;
  auto_synchronize true;
  ()

let human ((p, g) as s) =
  assert (p = B);
  match moves s with
  | [s'] when moves_ s = [] ->
      s' (* pas le choix *)
  | _ ->
      let rec ask () =
        let _st = wait_next_event [Button_down] in
        let x, y = mouse_pos () in
        Format.eprintf "click %d,%d@." x y;
        if ofs<=x && x<ofs+8*scale && ofs<=y && y<ofs+8*scale then
          let i, j = 7 - u_ y, u_ x in
          if can_play B g i j then play s (i, j) else ask ()
        else
          ask () in
      ask ()

let rec game s =
  draw s;
  if moves s <> [] then (
    let s = human s in
    draw s;
    if moves s <> [] then (
      printf "configuration:@\n%a@." print s;
      let s = computer s in
      printf "=> je joue@\n%a@." print s;
      game s
    )
  )

let () =
  open_graph " 800x800";
  try while true do
    game init;
    moveto 20 20;
    draw_string "rejouer ?";
    let c = read_key () in
    if c <> 'o' then raise Exit;
  done with Exit ->
    close_graph ()


(* vérification de l'interface *)

module GameOthello = Game.Make(struct

  type player_ = player type player = player_
  type state_ = state type state = state_

  let player = player
  let moves = moves

  let outcome = outcome

  let eval = eval

end)
