(**************************************************************************)
(*                                                                        *)
(*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       *)
(*  Kim Nguyen, Laurent Sartre                                            *)
(*                                                                        *)
(*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      *)
(*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 *)
(*                                                                        *)
(*  https://www.informatique-mpi.fr/                                      *)
(*                                                                        *)
(**************************************************************************)

(* Calcul propositionnel

   Les variables sont ici des entiers entre 1 et n.
   Cela permet notamment une représentation facile des littéraux
   (entier i ou -i) et une forme normale (CNF ou DNF) qui est
   directement au format DIMACS.
*)

open Format

type binop = And | Or | Implies

type fmla =
  | True
  | False
  | Var of int (* dans 1..n *)
  | Not of fmla
  | Bin of binop * fmla * fmla

let string_of_binop = function
  | And     -> "/\\"
  | Or      -> "\\/"
  | Implies -> "->"

let rec print fmt = function
  | True ->
      fprintf fmt "true"
  | False ->
      fprintf fmt "false"
  | Var i ->
      fprintf fmt "%d" i
  | Not f ->
      fprintf fmt "(@[not@ %a@])" print f
  | Bin (op, f1, f2) ->
      fprintf fmt "(@[%a %s@ %a@])" print f1 (string_of_binop op) print f2

let rec size = function
  | True | False | Var _ -> 1
  | Not f -> 1 + size f
  | Bin (_, f1, f2) -> 1 + size f1 + size f2

(* Quelques raccourcis et constructions

   Note : les fonctions comme `f_or` pourraient effectuer des
   simplifications, comme par exemple simplifier
   `Bin(Or, f, True)` en `f`. *)

let rec subst i v = function
  | Var j when i = j -> if v then True else False
  | Var _ | True | False as f -> f
  | Not f -> Not (subst i v f)
  | Bin (op, f1, f2) -> Bin (op, subst i v f1, subst i v f2)

let f_true = True
let f_false = False
let f_var i = Var i
let f_not f = Not f
let f_or f1 f2 = Bin (Or, f1, f2)
let f_and f1  f2 = Bin (And, f1, f2)
let f_implies f1  f2 = Bin (Implies, f1, f2)
let f_xor f1 f2 = f_or (f_and f1 (Not f2)) (f_and (Not f1) f2)
let f_equiv f1 f2 = f_and (f_implies f1 f2) (f_implies f2 f1)
let f_exists i f = f_or (subst i false f) (subst i true f)
let f_forall i f = f_and (subst i false f) (subst i true f)

(* Une valuation associe une valeur de vérité à chaque variable.
   La case 0 du tableau est inutilisée. *)

type valuation = bool array

let rec eval v = function
  | True -> true
  | False -> false
  | Var i -> assert (1 <= i && i < Array.length v); v.(i)
  | Not f -> not (eval v f)
  | Bin (And,     f1, f2) -> eval v f1 && eval v f2
  | Bin (Or,      f1, f2) -> eval v f1 || eval v f2
  | Bin (Implies, f1, f2) -> not (eval v f1) || eval v f2

(* Pour manipuler des tables de vérité ou des valuations, il peut être
   utile de commencer par déterminer l'ensemble des variables. *)

let varmax f =
  let rec varmax m = function
    | True | False -> m
    | Var i -> max i m
    | Not f -> varmax m f
    | Bin (_, f1, f2) -> varmax (varmax m f1) f2 in
  varmax 0 f

(* Table de vérité

   Affiche les 2^n lignes de la table, dans l'ordre lexicographique
   donné par F < T. *)

let print_truth_table fmt f =
  let nv = varmax f in
  let v = Array.make (nv + 1) false in
  let bool fmt b = fprintf fmt "%s" (if b then "T" else "F") in
  let rec fill i =
    if i = nv + 1 then (
      for x = 1 to nv do fprintf fmt "%a" bool v.(x) done;
      fprintf fmt " => %a@\n" bool (eval v f)
    ) else (
      v.(i) <- false; fill (i+1); v.(i) <- true;  fill (i+1)) in
  fill 1

(* Formes normales

   Un littéral `x` est un entier dans 1..n (variable x)
   ou dans -n..-1 (négation de la variable -x).

   Une clause `c` est une liste de littéraux, sans doublons,
   triée par ordre croissant de valeur absolue. *)

type literal = int
type clause = literal list
type kind = CNF | DNF
type nf = { kind: kind; nbvars: int; clauses: clause list }

let dimacs fmt nf =
  let nc = List.length nf.clauses in
  fprintf fmt "p %s %d %d@\n"
    (match nf.kind with CNF -> "cnf" | DNF -> "dnf") nf.nbvars nc;
  let literal x = fprintf fmt "%d " x in
  let clause c = List.iter literal c; fprintf fmt "0@\n" in
  List.iter clause nf.clauses

(* Forme normale conjonctive (CNF en anglais)

   Invariant : une CNF est
   - soit [] (true)
   - soit [[]] (false)
   - soit une liste non vide de clauses non vides
*)

(* la disjonction de deux clauses (exploite le caractère trié) *)
exception Tauto

let rec merge_or = function
  | [], c | c, [] -> c
  | (x1 :: s1 as c1), (x2 :: s2 as c2) ->
      if x1 = x2 then x1 :: merge_or (s1, s2) else
      if x1 = -x2 then raise Tauto else
      if abs x1 < abs x2 then x1 :: merge_or (s1, c2) else
      x2 :: merge_or (c1, s2)

(* toutes les disjonctions c1\/c2 avec c1 dans cl1 et c2 dans cl2 *)
let cnf_or cl1 cl2 = match cl1, cl2 with
  | [], _ | _, [] ->
      []
  | [[]], c | c, [[]] ->
      c
  | _ ->
      let combine11 c2 acc c1 =
        try let c = merge_or (c1, c2) in c :: acc with Tauto -> acc in
      let combine1 acc c2 = List.fold_left (combine11 c2) acc cl1 in
      List.fold_left combine1 [] cl2

let cnf_and cl1 cl2 =
  if cl1 = [[]] || cl2 = [[]] then [[]] else cl1 @ cl2

(* mise en forme normale conjonctive de `f` *)
let rec cnf = function
  | True -> []
  | False -> [[]]
  | Var x -> [[x]]
  | Not f -> cnf_neg f
  | Bin (And, f1, f2) -> cnf_and (cnf f1) (cnf f2)
  | Bin (Or, f1, f2) -> cnf_or (cnf f1) (cnf f2)
  | Bin (Implies, f1, f2) -> cnf_or (cnf_neg f1) (cnf f2)

(* mise en forme normale conjonctive de `non f` *)
and cnf_neg = function
  | True -> [[]]
  | False -> []
  | Var x -> [[-x]]
  | Not f -> cnf f
  | Bin (And, f1, f2) -> cnf_or (cnf_neg f1) (cnf_neg f2)
  | Bin (Or, f1, f2) -> cnf_and (cnf_neg f1) (cnf_neg f2)
  | Bin (Implies, f1, f2) -> cnf_and (cnf f1) (cnf_neg f2)

let cnf f =
  { kind = CNF; nbvars = varmax f; clauses = cnf f }

(* TODO : Forme normale disjonctive (DNF en anglais) *)

(* DPLL

   Pour une forme normale conjonctive `nf`, la fonction `dpll` renvoie
   - `Some v` si la formule est satisfiable, avec `v` une valuation qui convient
   - `None` si la formule n'est pas satisfiable. *)

exception Unsat

(* on a choisi le littéral x et on propage ; lève Unsat si contradiction *)
let propagate x cl =
  let rec simplify = function
    | [] -> []
    | y :: _ when y = x -> raise Exit
    | y :: c when y = -x -> simplify c
    | y :: c -> y :: simplify c in
  let clause cl c =
    match simplify c with
    | [] -> raise Unsat
    | c  -> c :: cl
    | exception Exit -> cl in
  List.fold_left clause [] cl

let dpll nf =
  if nf.kind <> CNF then invalid_arg "dpll";
  let n = nf.nbvars in
  let v = Array.make (n + 1) false in
  (* est-il possible de satisfaire une liste de clauses ? *)
  let rec backtrack = function
    | []      -> true
    | c :: cl -> List.exists (literal cl) c
  (* est-il possible de satisfaire le littéral x puis la liste cl ? *)
  and literal cl x =
    v.(abs x) <- x > 0;
    try let cl = propagate x cl in backtrack cl with Unsat -> false
  in
  if backtrack nf.clauses then Some v else None


(* MAX-2-SAT derandomized

   Pour une forme normale conjonctive `nf`, dont chaque clause contient
   au maximum deux littéraux, la fonction `max2sat` renvoie une valuation
   satisfaisant au moins la moitié des clauses (et donc au moins la moitié
   du nombre maximal de clauses satisfiables simultanément). *)

let random_max_sat nf =
  if nf.kind <> CNF then invalid_arg "random_max_sat";
  let n = nf.nbvars in
  Array.init (n + 1) (fun _ -> Random.bool ())

let max_ksat_approx nf =
  if nf.kind <> CNF then invalid_arg "max_ksat_approx";
  let n = nf.nbvars in
  let v = Array.make (n + 1) false in
  let max_k =
    List.fold_left max 0 (List.map List.length nf.clauses) in
  let b = 1 lsl max_k in
  for i = 1 to n do
    let expect_cl cl =
      let sat_literal x = abs x <= i && x > 0 = v.(abs x) in
      (* simplifications arithmétiques possibles *)
      if List.exists sat_literal cl then b
      else let cl' = List.filter (fun x -> abs x > i) cl in
           b - b / (1 lsl List.length cl')
    in
    let expect () = List.fold_left (+) 0
                      (List.map expect_cl nf.clauses) in
    let exp_f = expect () in
    v.(i) <- true;
    let exp_t = expect () in
    if exp_f > exp_t then v.(i) <- false
  done;
  v

let count_sat nf v =
  List.length (List.filter (List.exists (fun x -> x > 0 = v.(abs x))) nf.clauses)

let max_sat_exact nf =
  if nf.kind <> CNF then invalid_arg "max_sat_exact";
  let n = nf.nbvars in
  let v = Array.make (n + 1) false in
  let rec explore i =
    if i > n then
      count_sat nf v
    else
      (let a = explore (i+1) in
       v.(i) <- true;
       let b = explore (i+1) in
       v.(i) <- false;
       max a b)
  in
  explore 1

(* Réduction de SAT vers 3-SAT (Tseytin) *)
let tseytin f =
  let nv = ref (varmax f) in
  let new_var () = incr nv; !nv in
  let rec mk_clauses = function
    | Var z -> z, []
    | True  -> let x = new_var () in x, [[x]]
    | False -> let x = new_var () in x, [[-x]]
    | Not f -> let x = new_var () in
               let y, cls = mk_clauses f in
               x, [-x; -y] :: [x; y] :: cls
    | Bin (op, f1, f2) ->
       let x = new_var () in
       let y1, cls1 = mk_clauses f1 in
       let y2, cls2 = mk_clauses f2 in
       let cls = match op with
         | And     -> [-x; y1] :: [-x; y2] :: [x; -y1; -y2] :: cls1 @ cls2
         | Or      -> [-x; y1; y2] :: [x; -y1] :: [x; -y2] :: cls1 @ cls2
         | Implies -> [-x; -y1; y2] :: [x; y1] :: [x; -y2] :: cls1 @ cls2
       in
       x, cls
  in
  let x, cls = mk_clauses f in
  { kind = CNF; nbvars = !nv; clauses = [x] :: cls }


let polynomial_3sat (f: nf) : bool =
  failwith "the cake is a lie"

let polynomial_sat (f: fmla) : bool =
  let nf = tseytin f in
  polynomial_3sat nf

(* Faire autres réductions,
   y compris les passerelles entre décision/recherche/optim *)

(* Algorithme de Quine *)

let smart_not f = match f with
  | True  -> False
  | False -> True
  | _     -> Not f
let smart_and f1 f2 = match f1, f2 with
  | True, True -> True
  | False, _ | _, False -> False
  | _, _ -> Bin (And, f1, f2)
let smart_or f1 f2 = match f1, f2 with
  | True, _ | _, True -> True
  | False, False -> False
  | _, _ -> Bin (Or, f1, f2)
let smart_implies f1 f2 = match f1, f2 with
  | False, _ | _, True -> True
  | True, False -> False
  | _, _ -> Bin (Implies, f1, f2)

let rec simplify f = match f with
  | Var _ | True | False -> f
  | Not f            -> smart_not (simplify f)
  | Bin (op, f1, f2) -> let smart_op = match op with
                          | And     -> smart_and
                          | Or      -> smart_or
                          | Implies -> smart_implies
                        in
                        smart_op (simplify f1) (simplify f2)

let rec quine_sat f =
  match simplify f with
  | True  -> true
  | False -> false
  | f     -> let v = varmax f in
             quine_sat (subst v true f)
             || quine_sat (subst v false f)

(* MAXSAT Branch and Bound *)

let empty_clauses cnf =
  List.length (List.filter ((=) []) cnf)

let opposite_unit_clauses n cnf =
  let uc = Array.make (2 * n + 1) 0 in
  List.iter (fun c -> match c with
                      | x :: [] -> uc.(n + x) <- uc.(n + x) + 1
                      | _       -> ()) cnf;
  let o = ref 0 in
  for x = 1 to n do o := !o + min uc.(n - x) uc.(n + x) done;
  !o

let lb n cnf =
  empty_clauses cnf + opposite_unit_clauses n cnf

(* version de propagate qui conserve les clauses contradictoires *)
let propagate x cnf =
  let rec simplify_clause c = match c with
    | []                 -> []
    | y :: _ when y = x  -> raise Exit
    | y :: c when y = -x -> simplify_clause c
    | y :: c             -> y :: simplify_clause c in
  let fold_clause acc c =
    try simplify_clause c :: acc
    with Exit -> acc
  in
  List.fold_left fold_clause [] cnf

let maxsat nf =
  if nf.kind <> CNF then invalid_arg "maxsat_bnb";
  let n, nc = nf.nbvars, List.length nf.clauses in
  let ub = ref nc in
  let rec explore i cnf =
    if i > n then (
      let l = List.length cnf in
      if l < !ub then ub := l
    ) else
      if lb n cnf < !ub then (
        explore (i + 1) (propagate i cnf);
        explore (i + 1) (propagate (-i) cnf)
      ) (* pas de else *)
  in
  explore 1 nf.clauses;
  nc - !ub

let maxsat_bnb nf =
  if nf.kind <> CNF then invalid_arg "maxsat_bnb";
  let n, nc = nf.nbvars, List.length nf.clauses in
  let ub = ref nc in
  let rec explore i cnf =
    printf "prof %d : " (i - 1);
    if i > n then (
      let l = List.length cnf in
      if l < !ub then (ub := l; printf "new ub %d@." l)
      else printf "no better@."
    ) else
      if lb n cnf < !ub then (
        printf "explore@.";
        explore (i + 1) (propagate i cnf);
        explore (i + 1) (propagate (-i) cnf)
      ) else printf "cut@."
  in
  explore 1 nf.clauses;
  nc - !ub


(*
exception Sat
let simplify_cnf cnf x =
  let rec simplify_clause c = match c with
    | [] -> []
    | y :: _ when y = x -> raise Sat
    | y :: c when y = -x -> simplify_clause c
    | y :: c -> y :: simplify_clause c in
  let rec simplify cnf = match cnf with
    | [] -> []
    | c :: cnf ->
       (let cnf = simplify cnf in
        try let c = simplify_clause c in c :: cnf
        with Sat -> cnf)
  in
  simplify cnf
*)
(*
let maxsat_bnb nf =
  if nf.kind <> CNF then invalid_arg "maxsat_bnb";
  let n = nf.nbvars in
  let nc = List.length nf.clauses in
  let min_unsat = ref nc in
  let rec explore i cnf =
    if i > n then
      let unsat = count_empty cnf in
      (if unsat < !min_unsat then min_unsat := unsat)
    else
      if count_empty cnf < !min_unsat then
        (explore (i + 1) (propagate_cnf i cnf);
         explore (i + 1) (propagate_cnf (-i) cnf))
  in
  explore 1 nf.clauses;
  nc - !min_unsat
*)

(*
let maxsat_bnb_log nf =
  if nf.kind <> CNF then invalid_arg "maxsat_bnb";
  let n = nf.nbvars in
  let nc = List.length nf.clauses in
  (* let best_unsat = ref (count_sat nf (max_ksat_approx nf)) in *)
  let best_unsat = ref nc in

  let log = Array.make (n + 1) 0 in

  let rec explore i cnf =
    if i > n then
      let unsat = count_empty cnf (* + count_opposite n cnf *) in
      (if unsat < !best_unsat then (best_unsat := unsat;
                                    printf "new ub : %d@." unsat);
       log.(n) <- log.(n) + 1)
    else
      if count_empty cnf < !best_unsat then
        (explore (i + 1) (propagate_cnf i cnf);
         explore (i + 1) (propagate_cnf (-i) cnf))
      else
        log.(i) <- log.(i) + 1
  in
  explore 1 nf.clauses;
  nc - !best_unsat, log

let maxsat_bnb_log2 nf =
  if nf.kind <> CNF then invalid_arg "maxsat_bnb";
  let n = nf.nbvars in
  let nc = List.length nf.clauses in
  let best_unsat = ref (nc - count_sat nf (max_ksat_approx nf)) in

  let cut_log = Array.make (n + 1) 0 in
  let histo = Array.make (n + 2) 0 in

  let rec explore i cnf =
    histo.(i) <- histo.(i) + 1;
    if i > n then
      let unsat = count_empty cnf in
      (if unsat < !best_unsat then (best_unsat := unsat;
                                    printf "new ub : %d@." unsat);
       cut_log.(n) <- cut_log.(n) + 1)
    else
      if count_empty cnf + count_opposite n cnf < !best_unsat then
        (explore (i + 1) (propagate_cnf i cnf);
         explore (i + 1) (propagate_cnf (-i) cnf))
      else
        cut_log.(i) <- cut_log.(i) + 1
  in
  explore 1 nf.clauses;
  nc - !best_unsat, cut_log, histo
*)

(*
let maxsat_bnb_log3 nf =
  if nf.kind <> CNF then invalid_arg "maxsat_bnb";
  let n = nf.nbvars in
  let nc = List.length nf.clauses in
  let best_unsat = ref nc in

  let log = Array.make (n + 1) 0 in

  let rec explore i cnf =
    if i > n then
      let unsat = count_empty cnf in
      (if unsat < !best_unsat then (best_unsat := unsat;
                                    printf "new ub : %d@." unsat);
       log.(n) <- log.(n) + 1)
    else
      if count_empty cnf + count_opposite n cnf < !best_unsat then
        (explore (i + 1) (propagate_cnf i cnf);
         explore (i + 1) (propagate_cnf (-i) cnf))
      else
        log.(i) <- log.(i) + 1
  in
  explore 1 nf.clauses;
  nc - !best_unsat, log

let maxsat_bnb_log4 nf =
  if nf.kind <> CNF then invalid_arg "maxsat_bnb";
  let n = nf.nbvars in
  let nc = List.length nf.clauses in
  let best_unsat = ref (nc - count_sat nf (max_ksat_approx nf)) in

  let log = Array.make (n + 1) 0 in

  let rec explore i cnf =
    if i > n then
      let unsat = count_empty cnf in
      (if unsat < !best_unsat then (best_unsat := unsat;
                                    printf "new ub : %d@." unsat);
       log.(n) <- log.(n) + 1)
    else
      if count_empty cnf < !best_unsat then
        (explore (i + 1) (propagate_cnf i cnf);
         explore (i + 1) (propagate_cnf (-i) cnf))
      else
        log.(i) <- log.(i) + 1
  in
  explore 1 nf.clauses;
  nc - !best_unsat, log
*)
(* bilan : le calcul amélioré de la borne est plus important que
   la valeur de départ *)
