
(* Arbres rouge-noir, immuables. *)

type color = R | B

type ('k, 'v) rbt =
  | E
  | N of color * ('k, 'v) rbt * 'k * 'v * ('k, 'v) rbt
(* invariants
   (1) le père d'un noeud rouge n'est jamais rouge
   (2) le nombre de noeuds noirs le long d'un chemin de la racine à un
       sous-arbre vide est toujours le même *)

let empty : ('k, 'v) rbt =
  E

let is_empty (t: ('k, 'v) rbt) : bool =
  t = E

let rec find (k: 'k) (t: ('k, 'v) rbt) : 'v =
  match t with
  | E ->
      raise Not_found
  | N (_, l, k', v', r) ->
      if k < k' then find k l
      else if k > k' then find k r
      else v'

(* restaure les invariants si on a deux rouges consécutifs à gauche *)
let lbalance = function
  | (B, N (R, N (R, t1, k1, v1, t2), k2, v2, t3), k3, v3, t4)
  | (B, N (R, t1, k1, v1, N (R, t2, k2, v2, t3)), k3, v3, t4) ->
      N (R, N (B, t1, k1, v1, t2), k2, v2, N (B, t3, k3, v3, t4))
  | (c, l, k, v, r) -> N (c, l, k, v, r)

(* restaure les invariants si on a deux rouges consécutifs à droite *)
let rbalance = function
  | (B, t1, k1, v1, N (R, N (R, t2, k2, v2, t3), k3, v3, t4))
  | (B, t1, k1, v1, N (R, t2, k2, v2, N (R, t3, k3, v3, t4))) ->
      N (R, N (B, t1, k1, v1, t2), k2, v2, N (B, t3, k3, v3, t4))
  | (c, l, k, v, r) -> N (c, l, k, v, r)

let rec add (k: 'k) (v: 'v) (t: ('k, 'v) rbt) : ('k, 'v) rbt =
  match t with
  | E ->
      N (R, E, k, v, E)
  | N (c, l, k', v', r) ->
      if k < k' then lbalance (c, add k v l, k', v', r)
      else if k > k' then rbalance (c, l, k', v', add k v r)
      else N (c, l, k, v, r) (* on écrase la valeur *)

let add (k: 'k) (v: 'v) (t: ('k, 'v) rbt) : ('k, 'v) rbt =
  match add k v t with
  | E -> assert false
  | N (_, l, k', v', r) -> N (B, l, k', v', r)

(* exercice *)

let rec min_binding t = match t with
  | E -> raise Not_found
  | N (_, E, k, v, _) -> k, v
  | N (_, l, _, _, _) -> min_binding l

(* suppression *)

(* la fonction unbalanced_left (resp. unbalanced_right) répare
   l'invariant (2) lorsque la hauteur noire à gauche (resp. à droite)
   excède d'une unité la hauteur noire à droite (resp. à gauche).

   Le booléen renvoyé signale si la hauteur noire a diminué. *)
let unbalanced_left = function
  | N (R, N (B, t1, kx1,vx1, t2), kx2,vx2, t3) ->
      lbalance (B, N (R, t1, kx1,vx1, t2), kx2, vx2, t3), false
  | N (B, N (B, t1, kx1,vx1, t2), kx2,vx2, t3) ->
      lbalance (B, N (R, t1, kx1,vx1, t2), kx2, vx2, t3), true
  | N (B, N (R, t1, kx1,vx1, N (B, t2, kx2,vx2, t3)), kx3,vx3, t4) ->
      N (B, t1, kx1,vx1, lbalance (B, N (R, t2, kx2,vx2, t3), kx3, vx3, t4)), false
  | _ ->
      assert false

let unbalanced_right = function
  | N (R, t1, kx1,vx1, N (B, t2, kx2,vx2, t3)) ->
      rbalance (B, t1, kx1, vx1, N (R, t2, kx2,vx2, t3)), false
  | N (B, t1, kx1,vx1, N (B, t2, kx2,vx2, t3)) ->
      rbalance (B, t1, kx1, vx1, N (R, t2, kx2,vx2, t3)), true
  | N (B, t1, kx1,vx1, N (R, N (B, t2, kx2,vx2, t3), kx3,vx3, t4)) ->
      N (B, rbalance (B, t1, kx1, vx1, N (R, t2, kx2,vx2, t3)), kx3,vx3, t4), false
  | _ ->
      assert false

(* [remove_min t = (t',k,v,b)] extrait le minimum [k,v] de [t]. [t'] est le
    résultat et [b] indique si la hauteur noire a diminué *)
let rec remove_min = function
  | E ->
      assert false
  (* le minimum est atteint *)
  | N (B, E, kx,vx, E) ->
      E, kx,vx, true
  | N (B, E, kx,vx, N (R, l, ky,vy, r)) ->
      N (B, l, ky,vy, r), kx,vx, false
  | N (B, E, _,_, N (B,_,_,_,_)) ->
      assert false
  | N (R, E, kx,vx, r) ->
      r, kx,vx, false
  (* le minimum est récursivement extrait de [l] *)
  | N (c, l, kx,vx, r) ->
      let l',km,vm,d = remove_min l in
      let t = N (c, l', kx,vx, r) in
      if d then
        let t,d' = unbalanced_right t in t,km,vm,d'
      else
        t, km,vm, false

let blackify = function
  | N (R, l, kx,vx, r) -> N (B, l, kx,vx, r), false
  | s -> s, true

let remove kx s =
  (* [remove_aux t = (t',b)] supprime [kx] de [t] et indique avec [b]
     si la hauteur noire a diminué *)
  let rec remove_aux = function
    | E ->
        E, false
    | N (c, l, ky,vy, r) ->
        if kx < ky then
          let l',d = remove_aux l in
          let t = N (c, l', ky,vy, r) in
          if d then unbalanced_right t else t, false
        else if kx > ky then
          let r',d = remove_aux r in
          let t = N (c, l, ky,vy, r') in
          if d then unbalanced_left t else t, false
        else (* kx = ky *)
          (match r with
             | E ->
                 if c = R then l, false else blackify l
             | _ ->
                 let r',km,vm,d = remove_min r in
                 let t = N (c, l, km,vm, r') in
                 if d then unbalanced_left t else t, false)
  in
  let s',_ = remove_aux s in s'

(* exercice : tester la propriété d'arbre rouge et noir *)

exception Bad

let rec check_aux = function
  | E ->
      0
  | N (R, N (R,_,_,_,_), _, _, _)
  | N (R, _, _, _, N (R,_,_,_,_)) ->
      raise Bad
  | N (B, l, _, _, r) ->
      let h = check_aux l in
      if check_aux r <> h then raise Bad;
      h + 1
  | N (R, l, _, _, r) ->
      let h = check_aux l in
      if check_aux r <> h then raise Bad;
      h

let check s =
  try ignore (check_aux s); true with Bad -> false

(* notes : dans le code suivant,
   - `truncate neg_infinity` renvoie `0` donc `log2 0` renvoie `0`
     et donc pas besoin de cas particulier pour `n=0` (dans `log2` ou
     dans `of_array`)
   - `1 lsl -1` renvoie `0` donc là encore pas besoin de cas particulier
     pour `h=-1` daans `of_sub_array`
*)

let log2 n = truncate (log (float n) /. log 2.)

let rec of_sub_array a h lo hi =
  assert (0 <= lo && lo <= hi && hi <= Array.length a);
  assert (let n = hi - lo in 1 lsl h - 1 <= n && n < 1 lsl (h+1));
  if lo = hi then
    E
  else (
    let mid = lo + (hi - lo) / 2 in
    let k, v = a.(mid) in
    N ((if h = 0 then R else B),
       of_sub_array a (h-1) lo mid,
       k, v,
       of_sub_array a (h-1) (mid+1) hi)
  )

let of_array a =
  let n = Array.length a in
  of_sub_array a (log2 n) 0 n
