open Format

type re =
  | Empty
  | Epsilon
  | Char of char
  | Alt of re * re
  | Concat of re * re
  | Star of re

let prio = function
  | Empty | Epsilon | Char _ -> 10
  | Alt _ -> 5
  | Concat _ -> 6
  | Star _ -> 7

let pp fmt re =
  let rec pp_aux pr_par fmt re =
    let pr_self = prio re in
    let () = if pr_par > pr_self then fprintf fmt "(" in
    let () =
      match re with
      | Empty -> fprintf fmt "∅"
      | Epsilon -> fprintf fmt "ε"
      | Char (('|' | '(' | ')' | '*' | '\\') as c) -> fprintf fmt "\\%c" c
      | Char c -> fprintf fmt "%c" c
      | Alt (re1, re2) ->
          fprintf fmt "%a|%a" (pp_aux pr_self) re1 (pp_aux pr_self) re2
      | Concat (re1, re2) ->
          fprintf fmt "%a%a" (pp_aux pr_self) re1 (pp_aux pr_self) re2
      | Star re0 -> fprintf fmt "%a*" (pp_aux pr_self) re0
    in
    let () = if pr_par > pr_self then fprintf fmt ")" in
    ()
  in
  pp_aux 0 fmt re

let rec print fmt = function
  | Empty -> fprintf fmt "empty"
  | Epsilon -> fprintf fmt "epsilon"
  | Char ('0' .. 'z' as c) -> fprintf fmt "%c" c
  | Char c -> fprintf fmt "<%d>" (Char.code c)
  | Alt (r1, r2) -> fprintf fmt "(@[%a |@ %a@])" print r1 print r2
  | Concat (r1, r2) -> fprintf fmt "(%a%a)" print r1 print r2
  | Star r1 -> fprintf fmt "(%a*)" print r1

let rec has_epsilon = function
  | Empty -> false
  | Epsilon -> true
  | Char c -> false
  | Alt (re1, re2) -> has_epsilon re1 || has_epsilon re2
  | Concat (re1, re2) -> has_epsilon re1 && has_epsilon re2
  | Star _ -> true

let memo_table = Hashtbl.create 16

let memo re =
  try Hashtbl.find memo_table re
  with Not_found ->
    Hashtbl.add memo_table re re;
    re

let empty = Empty
let epsilon = Epsilon

let char c = memo (Char c)

let concat re1 re2 =
  match (re1, re2) with
  | Empty, _ | _, Empty -> Empty
  | Epsilon, e | e, Epsilon -> e
  | Star rre1, Star rre2 ->
      if rre1 == rre2 then re1 else memo (Concat (re1, re2))
  | _ -> memo (Concat (re1, re2))

let star = function
  | Empty | Epsilon -> Epsilon
  | Star _ as e -> e
  | e -> memo (Star e)

let alt re1 re2 =
  match (re1, re2) with
  | Empty, e | e, Empty -> e
  | Epsilon, e | e, Epsilon ->
      if has_epsilon e then e else memo (Alt (e, Epsilon))
  | _ when re1 == re2 -> re1
  | _ -> memo (Alt (re1, re2))