open Re

(* Dérivée de Brzozowski, version avec smart constructors *)
let rec derivative re c =
  match re with
  | Empty -> empty
  | Epsilon -> empty
  | Char c0 -> if c = c0 then epsilon else empty
  | Alt (re1, re2) -> alt (derivative re1 c) (derivative re2 c)
  | Concat (re1, re2) ->
      let cont = concat (derivative re1 c) re2 in
      if has_epsilon re1 then alt cont (derivative re2 c) else cont
  | Star re0 -> concat (derivative re0 c) re

let bmatch re s =
  let rec loop i n re =
    (* Format.eprintf "Position : %d, re: %a\n%!" i pp re; *)
    if i >= n then has_epsilon re else loop (i + 1) n (derivative re s.[i])
  in
  loop 0 (String.length s) re


(* Linéarisation.
   Les nouveaux caractères sont \001 ... \nnn.
   On fait l'hypothèse simplificatrice que 256 caractères seront
   suffisants pour la linéarisation. *)
let linearize r =
  let next = ref 0 in
  let m = Array.make 256 '\000' in
  (* nouveau -> ancien *)
  let add c =
    incr next;
    m.(!next) <- c;
    Char.chr !next
  in
  let rec lin = function
    | (Empty | Epsilon) as r -> r
    | Char c -> char (add c)
    | Alt (r1, r2) -> alt (lin r1) (lin r2)
    | Concat (r1, r2) -> concat (lin r1) (lin r2)
    | Star r1 -> star (lin r1)
  in
  let r = lin r in
  (!next, m, r)

(* Des ensembles de caractères.

   On a besoin de l'union, qui n'est pas dans `Rbt`.
   On utilise ici des tableaux de 256 booléens (Charset)
   ou bien le module `Set` d'OCaml. *)

module S = Charset
(* module S = Set.Make(Int) *)

let rec union l1 l2 =
     match l1, l2 with
      [], _ -> l2
      | _, [] -> l1
      | p1 :: ll1, p2 :: ll2 ->
        if p1 < p2 then p1 :: union ll1 l2 
        else if p1 = p2 then p1 :: union ll1 ll2
        else p2 :: union l1 ll2

(* Les premiers caractères des mots reconnus *)
let rec first r =
  match r with
  | Empty | Epsilon -> []
  | Char c -> [c]
  | Alt (r1, r2) ->union (first r1) (first r2)
  | Concat (r1, r2) ->
      union (first r1) (if has_epsilon r1 then first r2 else[])
  | Star r1 -> first r1

(* Les derniers caractères des mots reconnus *)
let rec last r =
   match r with
  | Empty | Epsilon -> []
  | Char c -> [c]
  | Alt (r1, r2) -> union (last r1) (last r2)
  | Concat (r1, r2) ->
      union (last r2) (if has_epsilon r2 then last r1 else [])
  | Star r1 -> last r1

(* Les suivants de `c` dans les mots reconnus. *)
let rec follow r c =
    match r with
  | Empty | Epsilon | Char _ -> []
  | Alt (r1, r2) ->
      union (follow r1 c) (follow r2 c)
  | Concat (r1, r2) ->
     union (follow r1 c)
      (union (follow r2 c)
           (if List.mem c (last r1) then first r2 else []))
  | Star r1 ->
      union (follow r1 c) (if List.mem c (last r1) then first r1 else [])

(* L'automate produit a pour structure principale

      0  -->  1  --> 2  -->  ...  -->  n

   où 1,2,...,n sont les caractères de la linéarisation.
   L'état initial est 0 (conforme à `Auto`).
   Les états terminaux sont donnés par `has_epsilon` et `last`.
   Les transitions sont fonnées par `follow`. *)

let berry_sethi r =
  let n, m, rl = linearize r in
  let f = first rl and l = last rl in
  let a = Auto.create (n + 1) in
  if has_epsilon rl then Auto.set_final a 0;
  for i = 1 to n do
    let ci = Char.chr i in
    if List.mem ci l then Auto.set_final a i;
    if List.mem ci f then Auto.add_trans a 0 m.(i) i;
    let s = follow rl ci in
    for j = 1 to n do
      if List.mem (Char.chr j) s then Auto.add_trans a i m.(j) j
    done
  done;
  a

(* Algorithme de Brzozowski et McCusley de réduction d'état *)

(** Type auxiliaire : pour chaque état de l'automate généralisé, on garde :
    - les prédécesseurs (annotés par une expression régulière)
    - les successeurs (annotés par une expression régulière)
*)
type state_info = {
  pred : (Auto.state, re) Hashtbl.t;
  succ : (Auto.state, re) Hashtbl.t;
}

(* Fonctions auxiliaires de manipulation de l'automate généralisé *)
let new_state_info () = { pred = Hashtbl.create 16; succ = Hashtbl.create 16 }

type gen_auto = state_info array

let create_gen_auto n = Array.init n (fun _ -> new_state_info ())

let update_info info_table s re =
  let old_re = try Hashtbl.find info_table s with Not_found -> empty in
  Hashtbl.replace info_table s (alt old_re re)

let update_trans_info gauto q s re =
  update_info gauto.(q).succ s re;
  update_info gauto.(s).pred q re

let normalize auto =
  (* Ajout d'un état initial et acceptant unique *)
  let auto_size = Auto.size auto in
  let nauto = Auto.create (auto_size + 2) in
  Auto.add_eps_trans nauto 0 1;
  Auto.set_final nauto (auto_size - 1);
  for q = 0 to auto_size - 1 do
    let trs = Auto.all_trans_opt auto q in
    List.iter
      (fun (chopt, s) -> Auto.add_trans_opt nauto (q + 1) chopt (s + 1))
      trs;
    if Auto.is_final auto q then Auto.add_eps_trans nauto (q + 1) (auto_size + 1)
  done;
  nauto

let gen_auto_from_auto nauto =
  let nauto_size = Auto.size nauto in
  let gauto = create_gen_auto nauto_size in
  for q = 0 to nauto_size - 1 do
    let trs = Auto.all_trans_opt nauto q in
    List.iter
      (fun (ch, s) ->
        let re = match ch with None -> epsilon | Some c -> char c in
        update_trans_info gauto q s re)
      trs
  done;
  gauto

let state_reduction auto =
  (* Normalisaton de l'automate *)
  let nauto = normalize auto in
  (* Création d'un graphe à partir de l'automate normalisé *)
  let gauto = gen_auto_from_auto nauto in
  (* Fusion des états *)
  for k = 1 to Auto.size nauto - 2 do
    let info = gauto.(k) in
    let re_self =
      star (try Hashtbl.find info.succ k with Not_found -> epsilon)
    in
    Hashtbl.iter
      (fun p re_pred ->
        Hashtbl.iter
          (fun q re_succ ->
            (* les états 1 <= p,q < k ont été supprimés *)
            if (not (1 <= p && p < k)) && not (1 <= q && q < k) then
              update_trans_info gauto p q
                (concat re_pred (concat re_self re_succ)))
          info.succ)
      info.pred
  done;
  (* L'expression régulière recherchée est dans succ de l'état initial vers
     l'état final *)
  Hashtbl.find gauto.(0).succ (Auto.size nauto - 1)

(* Construction de Thompson *)
let thompson re =
  let new_state =
    let s = ref ~-1 in
    fun () -> incr s; !s
  in
  let trans = ref [] in
  let add_trans t = trans := t :: !trans in
  let rec loop re =
    let in_s = new_state () in
    let out_s = new_state () in
    let () =
      match re with
      Empty -> ()
      | Epsilon -> add_trans (in_s, None, out_s)
      | Char c -> add_trans (in_s, Some c, out_s)
      | Alt (re1, re2) ->
        let in_s1, out_s1 = loop re1 in
        let in_s2, out_s2 = loop re2 in
        add_trans (in_s, None, in_s1);
        add_trans (in_s, None, in_s2);
        add_trans (out_s1, None, out_s);
        add_trans (out_s2, None, out_s)
      | Concat (re1, re2) ->
        let in_s1, out_s1 = loop re1 in
        let in_s2, out_s2 = loop re2 in
        add_trans (in_s, None, in_s1);
        add_trans (out_s1, None, in_s2);
        add_trans (out_s2, None, out_s)
      | Star re0 ->
        let in_s0, out_s0 = loop re0 in
        add_trans (out_s0, None, in_s0);
        add_trans (in_s, None, in_s0);
        add_trans (out_s0, None, out_s)
      in (in_s, out_s)
  in
  (* Comme on génère les nouveaux états en début de fonction,
     le premier état généré aura la valeur 0, et sera bien
     l'état initial de tout l'automate.
  *)
  let (initial, final) = loop re in
  let auto = Auto.create (new_state()) in
  Auto.set_final auto final;
  List.iter (fun (p,c,q) -> Auto.add_trans_opt auto p c q) !trans;
  auto
