(**************************************************************************)
(*                                                                        *)
(*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       *)
(*  Kim Nguyen, Laurent Sartre                                            *)
(*                                                                        *)
(*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      *)
(*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 *)
(*                                                                        *)
(*  https://www.informatique-mpi.fr/                                      *)
(*                                                                        *)
(**************************************************************************)

(* Expressions régulières *)

open Format

type regexp =
  | Empty
  | Epsilon
  | Char of char
  | Alt of regexp * regexp
  | Concat of regexp * regexp
  | Star of regexp

let rec print fmt = function
  | Empty -> fprintf fmt "empty"
  | Epsilon -> fprintf fmt "epsilon"
  | Char ('0'..'z' as c) -> fprintf fmt "%c" c
  | Char c -> fprintf fmt "<%d>" (Char.code c)
  | Alt (r1, r2) -> fprintf fmt "(@[%a |@ %a@])" print r1 print r2
  | Concat (r1, r2) -> fprintf fmt "(%a%a)" print r1 print r2
  | Star r1 -> fprintf fmt "(%a*)" print r1

(* le mot vide est-il reconnu ? *)
let rec null = function
  | Empty | Char _ -> false
  | Epsilon | Star _ -> true
  | Alt (r1, r2) -> null r1 || null r2
  | Concat (r1, r2) -> null r1 && null r2

(* Linéarisation.
   Les nouveaux caractères sont \001 ... \nnn.
   On fait l'hypothèse simplificatrice que 256 caractères seront
   suffisants pour la linéarisation. *)
let linearize r =
  let next = ref 0 in
  let m = Array.make 256 '\000' in (* nouveau -> ancien *)
  let add c = incr next; m.(!next) <- c; Char.chr !next in
  let rec lin = function
    | Empty | Epsilon as r -> r
    | Char c -> Char (add c)
    | Alt (r1, r2) -> Alt (lin r1, lin r2)
    | Concat (r1, r2) -> Concat (lin r1, lin r2)
    | Star r1 -> Star (lin r1) in
  let r = lin r in
  !next, m, r

(* Des ensembles de caractères.

   On a besoin de l'union, qui n'est pas dans `Rbt`.
   On utilise ici des tableaux de 256 booléens (Charset)
   ou bien le module `Set` d'OCaml. *)

module S = Charset
(* module S = Set.Make(Int) *)

(* Les premiers caractères des mots reconnus *)
let rec first = function
  | Empty | Epsilon -> S.empty
  | Char c -> S.singleton (Char.code c)
  | Alt (r1, r2) -> S.union (first r1) (first r2)
  | Concat (r1, r2) ->
      S.union (first r1) (if null r1 then first r2 else S.empty)
  | Star r1 -> first r1

(* Les derniers caractères des mots reconnus *)
let rec last = function
  | Empty | Epsilon -> S.empty
  | Char c -> S.singleton (Char.code c)
  | Alt (r1, r2) -> S.union (last r1) (last r2)
  | Concat (r1, r2) ->
      S.union (last r2) (if null r2 then last r1 else S.empty)
  | Star r1 -> last r1

(* Les suivants de `c` dans les mots reconnus. *)
let rec follow c = function
  | Empty | Epsilon | Char _ -> S.empty
  | Alt (r1, r2) -> S.union (follow c r1) (follow c r2)
  | Concat (r1, r2) ->
      S.union (follow c r1)
        (S.union (follow c r2)
           (if S.mem c (last r1) then first r2 else S.empty))
  | Star r1 ->
      S.union (follow c r1)
        (if S.mem c (last r1) then first r1  else S.empty)

(* L'automate produit a pour structure principale

      0  -->  1  --> 2  -->  ...  -->  n

   où 1,2,...,n sont les caractères de la linéarisation.
   L'état initial est 0 (conforme à `Auto`).
   Les états terminaux sont donnés par `null` et `last`.
   Les transitions sont fonnées par `follow`. *)

open Auto

let berry_sethi r =
  let n, m, rl = linearize r in
  let f = first rl and l = last rl in
  let a = create (n + 1) in
  if null rl then set_final a 0;
  for i = 1 to n do
    if S.mem i l then set_final a i;
    if S.mem i f then add_trans a 0 m.(i) i;
    let s = follow i rl in
    for j = 1 to n do if S.mem j s then add_trans a i m.(j) j done
  done;
  a

