
(* Tic-tac-toe

   X joue en premier
*)

type player = X | O

let other = function X -> O | O -> X

type state = player * string
  (* Un état est défini par
     - le joueur dont c'est le tour
     - une chaîne de longueur 9, de caractères 'X' 'O' '.' *)

let player (p, _) =
  p
let is_turn p s =
  p = player s

let init =
  X, "........."

let get_bit i j (_, g) =
  g.[3*i + j]

let can_play i j s =
  get_bit i j s = '.'

let char p =
  if p = X then 'X' else 'O'

let play i j ((p, g) as s) =
  assert (can_play i j s);
  (other p, String.init 9 (fun k -> if k = 3*i+j then char p else g.[k]))

(* Le joueur qui doit jouer a perdu la partie *)
let loose (p, g) =
  let c = char (other p) in
  let check w = Math.for_all 0 9 (fun k -> w.[k] = '0' || g.[k] = c) in
  List.exists check
    ["000000111"; "000111000"; "111000000"; "001001001";
     "010010010"; "100100100"; "001010100"; "100010001"; ]

let outcome s =
  if loose s then Some (other (player s)) else None

let moves = Memo.memo (fun _ s ->
  if loose s then [] else
  let rec fill acc i j =
    if i = 3 then acc else
    if j = 3 then fill acc (i+1) 0 else
    let acc = if can_play i j s then play i j s :: acc else acc in
    fill acc i (j+1) in
  fill [] 0 0
)

open Format

let print fmt s =
  let g = Array.make_matrix 3 3 ' ' in
  for i = 0 to 2 do for j = 0 to 2 do
    if get_bit i j s <> '.' then g.(i).(j) <- get_bit i j s
  done done;
  for i = 0 to 2 do
    for j = 0 to 2 do
      fprintf fmt "%c" g.(i).(j);
      if j < 2 then fprintf fmt "|"
    done;
    fprintf fmt "@\n"; if i < 2 then fprintf fmt "-+-+-@\n";
  done;
  fprintf fmt "  %c " (char (player s));
  if loose s then fprintf fmt "  looses@\n" else fprintf fmt "  plays@\n"

let of_string str =
  assert (String.length str = 9);
  let p = ref 0 in
  Math.fold_interval 0 9 () (fun () k ->
          match str.[k] with
          | 'X' -> incr p
          | 'O' -> decr p
          | _   -> ());
  assert (!p = 0 || !p = 1);
  ((if !p = 0 then X else O), str)

module S =
  Set.Make(struct type t = state let compare = Stdlib.compare end)

(* combien de configurations au total ? *)
let all =
  let states = Hashtbl.create 65536 in
  let e = ref 0 in
  let rec explore s =
    if not (Hashtbl.mem states s) then (
      Hashtbl.add states s ();
      (* printf "%a@." print s; *)
      let l = moves s in
      e := !e + List.length l;
      List.iter explore l
    ) in
  explore init;
  printf "%d sommets@." (Hashtbl.length states);
  printf "%d arcs@." !e;
  Hashtbl.fold (fun s () all -> S.add s all) states S.empty

(*
5478 sommets
16167 arcs
*)

(* et quelle taille si vu comme un arbre ? *)
let () =
  let count = Memo.memo (fun count s ->
    List.fold_left (fun c s' -> c + count s') 1 (moves s)
  ) in
  printf "arbre de taille %d@." (count init)

(*
arbre de taille 549946
*)

let final =
  S.filter (fun s -> moves s = []) all
let () =
  printf "%d états terminaux@." (S.cardinal final)
(* Positions gagnantes *)
let winX =
  S.fold
    (fun s att -> if loose s && is_turn O s then S.add s att else att)
    all S.empty
let () =
  printf "%d positions gagnantes pour X@." (S.cardinal winX)
  (* S.iter (fun s -> printf "  %a@." print s) winX *)
(* 626 *)
let winO =
  S.filter (fun s -> loose s && is_turn X s) all
let () =
  printf "%d positions gagnantes pour O@." (S.cardinal winO)
let draw =
  S.diff (S.diff final winX) winO
let () =
  printf "%d matchs nuls@." (S.cardinal draw)

(* Min-max

   Ici, on peut aller jusqu'au bout du calcul.
   Pour un jeu plus complexe, on limiterait la profondeur du calcul
   et on utiliserait une heuristique pour évaluer les positions
   sur lesquelles on arrête le calcul.
   Voir game.ml et othello.ml *)
let minmax p = Memo.memo (fun minmax s ->
  match moves s with
  | [] ->
      (match outcome s with
       | Some p' -> if p' = p then +1 else -1
       | None    -> 0)
  | s' :: l ->
      List.fold_left (if player s = p then max else min)
        (minmax s') (List.map minmax l)
)

let minmax = minmax X
let () = assert (minmax (of_string "XXOOXXXOO") = 0)
let () = assert (minmax (of_string "XX.OOOX..") = -1)
let () = assert (minmax (of_string "X..O.....") = 1)
let () = assert (minmax (of_string "XX.O.....") = 1)
let () = assert (minmax (of_string "XX.OO....") = 1)
let () = assert (minmax (of_string "XXXOO....") = 1)
let () = assert (minmax init = 0)

let () = S.iter (fun s -> assert (minmax s = 1)) winX

(* Calcul des attracteurs

   De manière générale, attr(p,i) est l'ensemble des états où le
   joueur p possède une stratégie gagnante en au plus i coups.

   Ci-dessous, on calcule les attracteurs pour le joueur X seulement,
   pour i croissant. *)
let attX =
  let rec compute att i =
    (* attention à ne pas utiliser directement List.for_all *)
    let for_all p = function [] -> false | l -> List.for_all p l in
    let add s acc =
      let test = if is_turn X s then List.exists else for_all in
      if test (fun s' -> S.mem s' att) (moves s) then S.add s acc else acc
    in
    let att' = S.fold add all att in
    if S.cardinal att' > S.cardinal att then compute att' (i+1) else att
  in
  compute winX 0

let () =
  printf "attracteur X : %d états@." (S.cardinal attX);
  (* S.iter (fun s -> printf "  %a@." print s) attX; *)
  ()
(* attracteur X de taille 2936 *)

let () = S.iter (fun s -> assert (minmax s = 1)) attX
let () = S.iter (fun s -> assert (minmax s <> 1)) (S.diff all attX)

let winning all p =
  let rec compute att i =
    (* attention à ne pas utiliser directement List.for_all *)
    let for_all p = function [] -> false | l -> List.for_all p l in
    let add s acc =
      let test = if player s = p then List.exists else for_all in
      if test (fun s' -> S.mem s' att) (moves s) then S.add s acc else acc
    in
    let att' = S.fold add all att in
    if S.cardinal att' > S.cardinal att then compute att' (i+1) else att
  in
  let start =
    S.filter
      (fun s -> moves s = [] && player s <> p && outcome s = Some p) all in
  compute start 0

let () =
  let attX = winning all X in
  printf "attracteur O : %d états@." (S.cardinal attX);
  let attO = winning all O in
  printf "attracteur O : %d états@." (S.cardinal attO);
  let other = S.diff (S.diff all attX) attO in
  printf "autres : %d états@." (S.cardinal other);
  (* S.iter (fun s -> printf "  %a@." print s) other; *)
  ()


(* vérification de l'interface *)

module GameTictactoe = Game.Make(struct

  type player_ = player type player = player_
  type state_ = state type state = state_

  let player = player
  let moves = moves

  let outcome = outcome

  let eval _p _s =
    assert false (*TODO*)

end)

(*
5478 sommets
16167 arcs
arbre de taille 549946
958 états terminaux
626 positions gagnantes pour X
316 positions gagnantes pour O
16 matchs nuls
attracteur X : 2936 états
attracteur O : 2936 états
attracteur O : 1474 états
autres : 1068 états
*)
