
(* Arbres préfixes (trie)

   réalisant ici des tableaux associatifs mutables dont les clés sont
   des chaînes. *)

type 'a trie = {
  mutable value: 'a option;
       branches: (char, 'a trie) Hashtbl.t;
}

let create () : 'a trie =
  { value = None; branches = Hashtbl.create 8; }

let contains (t: 'a trie) (s: string) : bool =
  let rec find t i =
    if i = String.length s then
      t.value <> None
    else
      try find (Hashtbl.find t.branches s.[i]) (i + 1)
      with Not_found -> false in
  find t 0

let get (t: 'a trie) (s: string) : 'a =
  let rec find t i =
    if i = String.length s then
      (match t.value with None -> raise Not_found | Some v -> v)
    else
      find (Hashtbl.find t.branches s.[i]) (i + 1)
  in
  find t 0

let put (t: 'a trie) (s: string) (v: 'a) : unit =
  let rec add t i =
    if i = String.length s then
      t.value <- Some v
    else
      let b =
        try
          Hashtbl.find t.branches s.[i]
        with Not_found ->
          let b = create () in
          Hashtbl.add t.branches s.[i] b;
          b
      in
      add b (i+1)
  in
  add t 0

(* remove naïf, qui ne nettoie pas l'arbre *)
let remove t s =
  let rec rmv t i =
    if i = String.length s then
      t.value <- None
    else
      rmv (Hashtbl.find t.branches s.[i]) (i+1);
  in
  try rmv t 0 with Not_found -> ()

(* cette fois en prenant soin de supprimer les branches vides en remontant *)
let is_empty t =
  t.value = None && Hashtbl.length t.branches = 0

let remove t s =
  let rec rmv t i =
    if i = String.length s then
      t.value <- None
    else
      let b = Hashtbl.find t.branches s.[i] in
      rmv b (i+1);
      if is_empty b then Hashtbl.remove t.branches s.[i]
  in
  try rmv t 0 with Not_found -> ()

let prefix_sub t s i0 =
  let rec find best t i =
    let best = match t.value with None -> best | Some v -> Some (i - i0, v) in
    if i = String.length s then
      best
    else
      try find best (Hashtbl.find t.branches s.[i]) (i + 1)
      with Not_found -> best in
  match find None t i0 with
  | None -> raise Not_found
  | Some (len, v) -> len, v

let prefix t s =
  prefix_sub t s 0

let pattern t ?(wildcard='_') pat =
  let n = String.length pat in
  let rec visit acc prefix t i =
    if i = n then
      match t.value with Some v -> (prefix, v) :: acc | None -> acc
    else if pat.[i] = wildcard then
      Hashtbl.fold (fun c t acc -> visit acc (prefix ^ String.make 1 c) t (i+1))
        t.branches acc
    else
      try visit acc (prefix ^ String.make 1 pat.[i])
            (Hashtbl.find t.branches pat.[i]) (i+1)
      with Not_found -> acc
  in
  visit [] "" t 0

let rec visit f prefix t =
  (match t.value with | Some v -> f prefix v | None -> ());
  Hashtbl.iter (fun c t -> visit f (prefix ^ String.make 1 c) t) t.branches

let iter f t =
  visit f "" t

let below f t s =
  let rec find t i =
    if i = String.length s then
      visit f s t
    else
      try find (Hashtbl.find t.branches s.[i]) (i + 1) with Not_found -> ()
  in
  find t 0
